import logging
import sys

import click
from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import install

# Human logging levels
log_levels = {
    "trace": logging.DEBUG,  # No deeper log value for Python
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}
CLICKVERBOSITY = click.Choice(list(log_levels.keys()))

# Rich console
console = Console()


def set_logging_config(verbosity: str):
    """A function to define the configuration of the logging module

    Args:
        verbosity:
            A valid verbosity level as defined in `log_levels`
    """
    level = log_levels[verbosity]
    rich_handler_kwargs = {
        "show_time": True,
        "show_level": True,
        "rich_tracebacks": True,
        "tracebacks_show_locals": True,
        "locals_max_string": None,
    }
    if verbosity in ("debug", "trace"):
        rich_handler_kwargs["show_path"] = True
        log_time_format = "%x %X"
    else:
        rich_handler_kwargs["show_path"] = False
        log_time_format = "%x %X"

    # Rich traceback color formatter
    error_console = Console(file=sys.stderr, log_time_format=log_time_format)
    install(console=error_console, width=None)

    # A first handler on the main console to have output synchronized with progress bar (which are also printed on the
    # main console)
    handlers = [RichHandler(level=level, console=console, **rich_handler_kwargs)]

    # Define the basic config
    logging.basicConfig(
        level=log_levels[verbosity], handlers=handlers, datefmt=log_time_format, format="{message}", style="{"
    )

    # Capture the warnings
    logging.captureWarnings(True)



import logging
import re
from typing import Any, Optional

import numpy as np
import numpy.linalg as nplin

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.utils import ConductorType, IsolationType, LineModel, LineType
from roseau.load_flow.utils.constants import CX, EPSILON_0, EPSILON_R, MU_0, OMEGA, PI, RHO, TAN_D
from roseau.load_flow.utils.units import Q_, ureg

logger = logging.getLogger(__name__)


class LineCharacteristics:
    """A class to store the line characteristics of lines"""

    _type_re = "|".join(x.code() for x in LineType)
    _material_re = "|".join(x.code() for x in ConductorType)
    _section_re = r"[1-9][0-9]*"
    _REGEXP_LINE_TYPE_NAME: re.Pattern = re.compile(
        rf"^({_type_re})_({_material_re})_{_section_re}$", flags=re.IGNORECASE
    )

    @ureg.wraps(None, (None, None, "ohm/km", "S/km"), strict=False)
    def __init__(self, type_name: str, z_line: np.ndarray, y_shunt: Optional[np.ndarray] = None):
        """LineCharacteristics constructor.

        Args:
            type_name:
                The name of the line characteristics.

            z_line:
                 The Z matrix of the line (Ohm/km).

            y_shunt:
                The Y matrix of the line (Siemens/km). This field is optional if the line has no shunt part.
        """
        self.type_name = type_name
        self.z_line = z_line
        self.y_shunt = y_shunt
        self._check_matrix()

    def __eq__(self, other):
        if not isinstance(other, LineCharacteristics):
            return NotImplemented
        return (
            self.type_name == other.type_name
            and self.z_line.shape == other.z_line.shape
            and np.allclose(self.z_line, other.z_line)
            and (
                (self.y_shunt is None and other.y_shunt is None)
                or (
                    self.y_shunt is not None
                    and other.y_shunt is not None
                    and self.y_shunt.shape == other.y_shunt.shape
                    and np.allclose(self.y_shunt, other.y_shunt)
                )
            )
        )

    @classmethod
    @ureg.wraps(
        None,
        (
            None,
            None,
            None,
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "S/km",
            "S/km",
            "S/km",
            "S/km",
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "S/km",
            "S/km",
        ),
        strict=False,
    )
    def from_sym(
        cls,
        type_name: str,
        model: LineModel,
        r0: float,
        x0: float,
        r1: float,
        x1: float,
        g0: float,
        b0: float,
        g1: float,
        b1: float,
        rn: Optional[float] = None,
        xn: Optional[float] = None,
        xpn: Optional[float] = None,
        bn: Optional[float] = None,
        bpn: Optional[float] = None,
    ):
        """Create line characteristics from sym model.

        Args:
            type_name:
                The name of the line characteristics.

            model:
                The required model. It can be SYM or SYM_NEUTRAL. Be careful, it can be downgraded...

            r0:
                Resistance - zero sequence (ohms/km)

            x0:
                reactance - zero sequence  (ohms/km)

            r1:
                resistance - direct sequence (ohms/km)

            x1:
                reactance - direct sequence (ohms/km)

            g0:
                Conductance - zero sequence (Siemens/km)

            b0:
                Conductance - direct sequence (Siemens/km)

            g1:
                Susceptance - zero sequence (Siemens/km)

            b1:
                Susceptance - direct sequence (Siemens/km)

            rn:
                Neutral resistance (ohms/km)

            xn:
                Neutral reactance  (ohms/km)

            xpn:
                Phase to neutral reactance  (ohms/km)

            bn:
                Neutral susceptance (siemens/km)

            bpn:
                Phase to neutral susceptance (siemens/km)

        Returns:
            The created line characteristics.
        """
        z_line, y_shunt, model = cls._sym_to_zy(
            type_name=type_name,
            model=model,
            r0=r0,
            x0=x0,
            r1=r1,
            x1=x1,
            g0=g0,
            b0=b0,
            g1=g1,
            b1=b1,
            rn=rn,
            xn=xn,
            xpn=xpn,
            bn=bn,
            bpn=bpn,
        )
        return cls(type_name=type_name, z_line=z_line, y_shunt=y_shunt)

    @staticmethod
    @ureg.wraps(
        ("ohm/km", "S/km", None),
        (
            None,
            None,
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "S/km",
            "S/km",
            "S/km",
            "S/km",
            "ohm/km",
            "ohm/km",
            "ohm/km",
            "S/km",
            "S/km",
        ),
        strict=False,
    )
    def _sym_to_zy(
        type_name: str,
        model: LineModel,
        r0: float,
        x0: float,
        r1: float,
        x1: float,
        g0: float,
        b0: float,
        g1: float,
        b1: float,
        rn: Optional[float] = None,
        xn: Optional[float] = None,
        xpn: Optional[float] = None,
        bn: Optional[float] = None,
        bpn: Optional[float] = None,
    ) -> tuple[np.ndarray, np.ndarray, LineModel]:
        """Create impedance and admittance matrix from a symmetrical model.

        Args:
            type_name:
                The name of this type.

            model:
                The required model. It can be SYM or SYM_NEUTRAL. Be careful, it can be downgraded...

            r0:
                Resistance - zero sequence (ohms/km)

            x0:
                reactance - zero sequence  (ohms/km)

            r1:
                resistance - direct sequence (ohms/km)

            x1:
                reactance - direct sequence (ohms/km)

            g0:
                Conductance - zero sequence (Siemens/km)

            b0:
                Conductance - direct sequence (Siemens/km)

            g1:
                Susceptance - zero sequence (Siemens/km)

            b1:
                Susceptance - direct sequence (Siemens/km)

            rn:
                Neutral resistance (ohms/km)

            xn:
                Neutral reactance  (ohms/km)

            xpn:
                Phase to neutral reactance  (ohms/km)

            bn:
                Neutral susceptance (siemens/km)

            bpn:
                Phase to neutral susceptance (siemens/km)

        Returns:
            The impedance and admittance matrices and the line model. The line model may be downgraded from
            SYM_NEUTRAL to SYM if the model of the neutral is not possible.
        """
        # Extract the data
        z0 = r0 + 1j * x0
        z1 = r1 + 1j * x1
        y0 = g0 + 1j * b0
        y1 = g1 + 1j * b1

        # Two possible choices. The first one is the best but sometimes PwF data forces us to choose the second one
        for choice in (0, 1):
            if choice == 0:
                # We trust the manual !!! can gives singular matrix !!!
                zs = (z0 + 2 * z1) / 3  # Serie impedance (ohms/km)
                zm = (z0 - z1) / 3  # Mutual impedance (ohms/km)

                ys = (y0 + 2 * y1) / 3  # Serie shunt admittance (siemens/km)
                ym = (y0 - y1) / 3  # Mutual shunt admittance (siemens/km)
            else:
                # Do not read the manual, it is useless: in pwf we trust
                zs = r1 + 1j * x1  # Serie impedance (ohms/km)
                zm = 0 + 0j  # Mutual impedance (ohms/km)

                ys = g1 + 1j * b1  # Serie shunt admittance (siemens/km)
                ym = 0 + 0j  # Mutual shunt admittance (siemens/km)

            if model == LineModel.SYM_NEUTRAL:
                # Add the neutral
                # Build the complex
                zn = rn + 1j * xn  # Neutral serie impedance (ohm/km)
                zpn = xpn * 1j  # Phase-to-neutral serie impedance (ohm/km)
                yn = bn * 1j  # Neutral shunt admittance (Siemens/km)
                ypn = bpn * 1j  # Phase to neutral shunt admittance (Siemens/km)

                if zpn == 0 and zn == 0:
                    logger.warning(
                        f"The low voltages line {type_name!r} does not have neutral elements. It will model as an "
                        f"medium voltages line."
                    )
                    z_line = np.array([[zs, zm, zm], [zm, zs, zm], [zm, zm, zs]], dtype=complex)

                    y_shunt = np.array([[ys, ym, ym], [ym, ys, ym], [ym, ym, ys]], dtype=complex)
                    # We downgrade the model to sym
                    model = LineModel.SYM
                else:

                    z_line = np.array(
                        [[zs, zm, zm, zpn], [zm, zs, zm, zpn], [zm, zm, zs, zpn], [zpn, zpn, zpn, zn]],
                        dtype=complex,
                    )

                    y_shunt = np.array(
                        [[ys, ym, ym, ypn], [ym, ys, ym, ypn], [ym, ym, ys, ypn], [ypn, ypn, ypn, yn]],
                        dtype=complex,
                    )

            else:
                assert model == LineModel.SYM
                z_line = np.array([[zs, zm, zm], [zm, zs, zm], [zm, zm, zs]], dtype=complex)

                y_shunt = np.array([[ys, ym, ym], [ym, ys, ym], [ym, ym, ys]], dtype=complex)

            # Check the validity of the resulting matrices
            det_z = nplin.det(z_line)
            if abs(det_z) == 0:
                if choice == 0:
                    # Warn the user that the PwF data are bad...
                    logger.warning(
                        f"The conversion of the symmetric model of {type_name!r} to its matrix model reached invalid "
                        f"line impedance matrix... It is often the case with line models coming from "
                        "PowerFactory. We pass to a 'degraded' model of lines to handle the provided data."
                    )
                    # Go to choice == 1
                else:
                    assert choice == 1
                    msg = (
                        f"The data provided for the line type {type_name!r} does not allow us to create a valid shunt "
                        f"admittance and line impedance."
                    )
                    logger.error(msg)
                    raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SHUNT_VALUE)
            else:
                # Break: the current choice is good!
                break

        return z_line, y_shunt, model

    @classmethod
    @ureg.wraps(
        None,
        (None, None, None, None, None, "mm**2", "mm**2", "m", "m"),
        strict=False,
    )
    def from_lv_exact(
        cls,
        type_name: str,
        line_type: LineType,
        conductor_type: ConductorType,
        isolation_type: IsolationType,
        section: float,
        section_neutral: float,
        height: float,
        external_diameter: float,
    ) -> "LineCharacteristics":
        """Create line characteristics from LV exact model.

        Args:
            type_name:
                The name of the line characteristics.

            line_type:
                Overhead or underground.

            conductor_type:
                Type of the conductor

            isolation_type:
                Type of insulator.

            section:
                Surface of the phases (mm²).

            section_neutral:
                Surface of the neutral (mm²).

            height:
                 Height of the line (m).

            external_diameter:
                External diameter of the wire (m).

        Returns:
            The created line characteristics.

        TODO: Documentation on the line data
        """
        z_line, y_shunt, model = cls._lv_exact_to_zy(
            type_name=type_name,
            line_type=line_type,
            conductor_type=conductor_type,
            insulator_type=isolation_type,
            section=section,
            section_neutral=section_neutral,
            height=height,
            external_diameter=external_diameter,
        )
        return cls(type_name=type_name, z_line=z_line, y_shunt=y_shunt)

    @staticmethod
    @ureg.wraps(
        ("ohm/km", "S/km", None),
        (None, None, None, None, "mm**2", "mm**2", "m", "m"),
        strict=False,
    )
    def _lv_exact_to_zy(
        type_name: str,
        line_type: LineType,
        conductor_type: ConductorType,
        insulator_type: IsolationType,
        section: float,
        section_neutral: float,
        height: float,
        external_diameter: float,
    ) -> tuple[np.ndarray, np.ndarray, LineModel]:
        """Create impedance and admittance matrix from a LV exact model.

        Args:
            type_name:
                The name of this type.

            line_type:
                Overhead or underground.

            conductor_type:
                Type of the conductor

            insulator_type:
                Type of insulator.

            section:
                Surface of the phases (mm²).

            section_neutral:
                Surface of the neutral (mm²).

            height:
                 Height of the line (m).

            external_diameter:
                External diameter of the wire (m).

        Returns:
            The impedance and admittance matrices and the line model.
        """
        # dpp = data["dpp"]  # Distance phase to phase (m)
        # dpn = data["dpn"]  # Distance phase to neutral (m)
        # dsh = data["dsh"]  # Diameter of the sheath (mm)

        # Geometric configuration
        if line_type == LineType.OVERHEAD or line_type == LineType.TWISTED:
            coord = Q_(
                np.array(
                    [
                        [
                            -np.sqrt(3) / 8 * external_diameter,
                            height + external_diameter / 8,
                            -np.sqrt(3) / 8 * external_diameter,
                            -height - external_diameter / 8,
                        ],
                        [
                            np.sqrt(3) / 8 * external_diameter,
                            height + external_diameter / 8,
                            np.sqrt(3) / 8 * external_diameter,
                            -height - external_diameter / 8,
                        ],
                        [0, height - external_diameter / 4, 0, -height + external_diameter / 4],
                        [0, height, 0, -height],
                    ]
                ),
                "m",
            )
            epsilon = EPSILON_0
        elif line_type == LineType.UNDERGROUND:
            coord = Q_(
                np.array(
                    [
                        [
                            -np.sqrt(2) / 8 * external_diameter,
                            height - np.sqrt(2) / 8 * external_diameter,
                            -np.sqrt(2) * 3 / 8 * external_diameter,
                            height - np.sqrt(2) * 3 / 8 * external_diameter,
                        ],
                        [
                            np.sqrt(2) / 8 * external_diameter,
                            height - np.sqrt(2) / 8 * external_diameter,
                            np.sqrt(2) * 3 / 8 * external_diameter,
                            height - np.sqrt(2) * 3 / 8 * external_diameter,
                        ],
                        [
                            np.sqrt(2) / 8 * external_diameter,
                            height + np.sqrt(2) / 8 * external_diameter,
                            np.sqrt(2) * 3 / 8 * external_diameter,
                            height + np.sqrt(2) * 3 / 8 * external_diameter,
                        ],
                        [
                            -np.sqrt(2) / 8 * external_diameter,
                            height + np.sqrt(2) / 8 * external_diameter,
                            -np.sqrt(2) * 3 / 8 * external_diameter,
                            height + np.sqrt(2) * 3 / 8 * external_diameter,
                        ],
                    ]
                ),
                "m",
            )
            epsilon = EPSILON_0 * EPSILON_R[insulator_type]
        else:
            msg = f"The line type of the line {type_name!r} is unknown. It should have been filled in the reading."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_LINE_TYPE)

        # Electrical parameters
        sections = Q_([section, section, section, section_neutral], "mm**2")  # surfaces (m2)
        radius = Q_(np.zeros(4, dtype=float), "m")  # radius (m)
        gmr = Q_(np.zeros(4, dtype=float), "m")  # geometric mean radius (m)
        # d = Q_(np.zeros((4, 4), dtype=float), "m")  # distance between projections of two wires (m)
        distance = Q_(np.zeros((4, 4), dtype=float), "m")  # distance between two wires (m)
        distance_prim = Q_(np.zeros((4, 4), dtype=float), "m")  # distance between a wire and the image of another
        # wire (m)
        r = Q_(np.zeros((4, 4), dtype=float), "ohm/km")  # resistance (ohm/km)
        inductance = Q_(np.zeros((4, 4), dtype=float), "H/km")  # inductance (H/km)
        lambdas = Q_(np.zeros((4, 4), dtype=float), "m/F")  # potential coefficient (m/F)
        for i in range(4):
            radius[i] = np.sqrt(sections[i] / PI)
            gmr[i] = radius[i].to("m") * np.exp(-0.25)
            r[i, i] = RHO[conductor_type] / sections[i]
            for j in range(4):
                # d[i, j] = abs(coord[i][0] - coord[j][0])
                distance[i, j] = np.sqrt((coord[i][0] - coord[j][0]) ** 2 + (coord[i][1] - coord[j][1]) ** 2)
                distance_prim[i, j] = np.sqrt((coord[i][0] - coord[j][2]) ** 2 + (coord[i][1] - coord[j][3]) ** 2)
                if j != i:
                    inductance[i, j] = MU_0 / (2 * PI) * np.log(Q_(1, "m") / distance[i, j])
                    inductance[j, i] = inductance[i, j]
                    lambdas[i, j] = 1 / (2 * PI * epsilon) * np.log(distance_prim[i, j] / distance[i, j])
                    lambdas[j, i] = lambdas[i, j]
            inductance[i, i] = MU_0 / (2 * PI) * np.log(Q_(1, "m") / gmr[i])
            lambdas[i, i] = 1 / (2 * PI * epsilon) * np.log(distance_prim[i, i] / radius[i])
        lambda_inv = Q_(nplin.inv(lambdas.magnitude), 1 / lambdas.units).to("F/km")  # capacities (F/km)
        c = Q_(np.zeros((4, 4), dtype=float), "F/km")  # capacities (F/km)
        g = Q_(np.zeros((4, 4), dtype=float), "S/km")  # conductance (S/km)
        for i in range(4):
            c[i, i] = lambda_inv[i, i]
            for j in range(4):
                if j != i:
                    c[i, i] -= lambda_inv[i, j]
                    c[i, j] = -lambda_inv[i, j]
            g[i, i] = TAN_D[insulator_type] * c[i, i] * OMEGA

        z_line = r + inductance * OMEGA * 1j
        y = g + c * OMEGA * 1j

        y_shunt = Q_(np.zeros((4, 4), dtype=complex), "S/km")
        for i in range(4):
            for k in range(4):
                y_shunt[i, i] += y[i, k]
            for j in range(4):
                if i != j:
                    y_shunt[i, j] = -y[i, j]

        return z_line, y_shunt, LineModel.LV_EXACT

    @classmethod
    def from_name_lv(
        cls,
        name: str,
        section_neutral: Optional[float] = None,
        height: Optional[float] = None,
        external_diameter: Optional[float] = None,
    ) -> "LineCharacteristics":
        """Method to get the electrical characteristics of a LV line from its canonical name.
        Some hypothesis will be made: the section of the neutral is the same as the other sections, the height and
        external diameter are pre-defined, and the isolation is PVC.

        Args:
            name:
                The name of the line the characteristics must be computed. Eg. "S_AL_150".

            section_neutral:
                Surface of the neutral (mm²). If None it will be the same as the section of the other phases.

            height:
                 Height of the line (m). If None a default value will be used.

            external_diameter:
                External diameter of the wire (mm). If None a default value will be used.

        Returns:
            The corresponding line characteristics.
        """
        match: re.Match = cls._REGEXP_LINE_TYPE_NAME.fullmatch(string=name)
        if not match:
            msg = f"The line type name does not follow the syntax rule. {name!r} was provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TYPE_NAME_SYNTAX)

        # Check the user input and retrieve enumerated types
        line_type, conductor_type, section = name.split("_")
        line_type = LineType.from_string(string=line_type)
        conductor_type = ConductorType.from_string(conductor_type)
        isolation_type = IsolationType.PVC

        section = float(section)

        if section_neutral is None:
            section_neutral = section
        if height is None:
            height = Q_(-1.5, "m") if line_type == LineType.UNDERGROUND else Q_(10.0, "m")
        if external_diameter is None:
            external_diameter = Q_(40, "mm")

        return cls.from_lv_exact(
            type_name=name,
            line_type=line_type,
            conductor_type=conductor_type,
            isolation_type=isolation_type,
            section=section,
            section_neutral=section_neutral,
            height=height,
            external_diameter=external_diameter,
        )

    @classmethod
    def from_name_mv(cls, name: str):
        """Method to get the electrical characteristics of a MV line from its canonical name.

        Args:
            name:
                The name of the line the characteristics must be computed. Eg. "S_AL_150".

        Returns:
            The corresponding line characteristics.
        """
        match: re.Match = cls._REGEXP_LINE_TYPE_NAME.fullmatch(string=name)
        if not match:
            msg = f"The line type name does not follow the syntax rule. {name!r} was provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TYPE_NAME_SYNTAX)

        # Check the user input and retrieve enumerated types
        line_type, conductor_type, section = name.split("_")
        line_type = LineType.from_string(string=line_type)
        conductor_type = ConductorType.from_string(conductor_type)
        section = Q_(float(section), "mm**2")

        r = RHO[conductor_type] / section
        x = CX[line_type]
        if type == LineType.OVERHEAD:
            c_b1 = Q_(50, "µF/km")
            c_b2 = Q_(0, "µF/(km*mm**2)")
        elif type == LineType.TWISTED:
            # Twisted line
            c_b1 = Q_(1750, "µF/km")
            c_b2 = Q_(5, "µF/(km*mm**2)")
        else:
            if section <= Q_(50, "mm**2"):
                c_b1 = Q_(1120, "µF/km")
                c_b2 = Q_(33, "µF/(km*mm**2)")
            else:
                c_b1 = Q_(2240, "µF/km")
                c_b2 = Q_(15, "µF/(km*mm**2)")
        b = (c_b1 + c_b2 * section) * 1e-4 * OMEGA
        b = b.to("S/km")

        z_line = (r + x * 1j) * np.eye(3)  # in ohms/km
        y_shunt = b * 1j * np.eye(3)  # in siemens/km
        return cls(type_name=name, z_line=z_line, y_shunt=y_shunt)

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        """Line characteristics constructor from dict.

        Args:
            data:
                The dictionary data of the line characteristics.

        Returns:
            The created line characteristics.
        """
        type_name = data["name"]
        model = LineModel.from_string(data["model"])
        if model == LineModel.LV_EXACT:
            return cls.from_lv_exact(type_name=type_name, **data)
        elif model in (LineModel.SYM_NEUTRAL, LineModel.SYM):
            return cls.from_sym(type_name=type_name, **data)
        elif model in (LineModel.ZY_NEUTRAL, LineModel.ZY, LineModel.Z, LineModel.Z_NEUTRAL):
            z_line = np.asarray(data["z_line"][0]) + 1j * np.asarray(data["z_line"][1])
            if "y_shunt" in data:
                y_shunt = np.asarray(data["y_shunt"][0]) + 1j * np.asarray(data["y_shunt"][1])
            else:
                y_shunt = None
            return cls(type_name=type_name, z_line=z_line, y_shunt=y_shunt)
        else:
            msg = f"The line {type_name!r} has an unknown model... We can do nothing."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_LINE_MODEL)

    def to_dict(self) -> dict[str, Any]:
        """Return the line characteristics information as a dictionary format."""
        res = {
            "name": self.type_name,
            "model": "zy_neutral" if self.y_shunt is not None else "z_neutral",
            "z_line": [self.z_line.real.tolist(), self.z_line.imag.tolist()],
        }
        if self.y_shunt is not None:
            res["y_shunt"] = [self.y_shunt.real.tolist(), self.y_shunt.imag.tolist()]
        return res

    def _check_matrix(self):
        """Check the coefficients of the matrix."""
        for matrix, matrix_name, code in [
            (self.y_shunt, "y_shunt", RoseauLoadFlowExceptionCode.BAD_Y_SHUNT_SHAPE),
            (self.z_line, "z_line", RoseauLoadFlowExceptionCode.BAD_Z_LINE_SHAPE),
        ]:
            if matrix_name == "y_shunt" and self.y_shunt is None:
                continue
            if matrix.shape[0] != matrix.shape[1]:
                msg = f"Incorrect {matrix_name} dimensions for line characteristics {self.type_name!r}: {matrix.shape}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=code)

        # Check of the coefficients value
        for matrix, matrix_name, code in [
            (self.z_line, "line impedance", RoseauLoadFlowExceptionCode.BAD_Z_LINE_VALUE),
            (self.y_shunt, "shunt admittance", RoseauLoadFlowExceptionCode.BAD_Y_SHUNT_VALUE),
        ]:
            if matrix_name == "shunt admittance" and self.y_shunt is None:
                continue
            # Check that the off-diagonal element have a zero real part
            off_diagonal_elements = matrix[~np.eye(*matrix.shape, dtype=np.bool_)]
            if not np.allclose(off_diagonal_elements.real, 0):
                msg = (
                    f"The {matrix_name} matrix of {self.type_name!r} has off-diagonal elements with a non-zero real "
                    f"part."
                )
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=code)
            # Check that the real coefficients are positive
            if (matrix.real < 0.0).any():
                msg = f"Some real part coefficients of the {matrix_name} matrix of {self.type_name!r} are negative..."
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=code)



import logging
from typing import Any

import numpy as np

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.utils import TransformerType, ureg

logger = logging.getLogger(__name__)


class TransformerCharacteristics:
    """A class to store the characteristics of the transformers."""

    @ureg.wraps(None, (None, None, None, "V", "V", "VA", "W", None, "W", None), strict=False)
    def __init__(
        self,
        type_name: str,
        windings: str,
        uhv: float,
        ulv: float,
        sn: float,
        p0: float,
        i0: float,
        psc: float,
        vsc: float,
    ):
        """TransformerCharacteristics constructor.

        Args:
            type_name:
                The name of the transformer type

            windings:
                The type of windings such as "Dyn11"

            uhv:
                Phase-to-phase nominal voltages of the high voltages side (V)

            ulv:
                Phase-to-phase nominal voltages of the low voltages side (V)

            sn:
                The nominal voltages of the transformer (VA)

            p0:
                Losses during off-load test (W)

            i0:
                Current during off-load test (%)

            psc:
                Losses during short circuit test (W)

            vsc:
                Voltages on LV side during short circuit test (%)
        """
        self.type_name = type_name
        self.sn = sn
        self.uhv = uhv
        self.ulv = ulv
        self.i0 = i0
        self.p0 = p0
        self.psc = psc
        self.vsc = vsc
        self.windings = windings
        self.winding1, self.winding2, self.phase_displacement = TransformerType.extract_windings(string=windings)

        # Check
        if uhv <= ulv:
            msg = (
                f"The transformer type {type_name!r} has a high voltages lower or equal than the low voltages: uhv="
                f"{uhv:.2f} V and ulv={ulv:.2f} V"
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_VOLTAGES)
        if i0 > 1.0 or i0 < 0.0:
            msg = (
                f"The transformer type {type_name!r} has a current during off-load test i0={i0}. It is a percentage "
                f"that should be between 0 and 1."
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_PARAMETERS)
        if vsc > 1.0 or vsc < 0.0:
            msg = (
                f"The transformer type {type_name!r} has a voltages on LV side during short circuit test vsc={vsc}. "
                f"It is a percentage that should be between 0 and 1."
            )
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_PARAMETERS)

    def __eq__(self, other):
        if not isinstance(other, TransformerCharacteristics):
            return NotImplemented
        else:
            return (
                self.type_name == other.type_name
                and self.windings == other.windings
                and np.isclose(self.sn, other.sn)
                and np.isclose(self.p0, other.p0)
                and np.isclose(self.i0, other.i0)
                and np.isclose(self.uhv, other.uhv)
                and np.isclose(self.ulv, other.ulv)
                and np.isclose(self.psc, self.psc)
                and np.isclose(self.vsc, other.vsc)
            )

    @classmethod
    def from_name(cls, name: str, windings: str) -> "TransformerCharacteristics":
        """TransformerCharacteristics constructor from name.

        Args:
            name:
                The name of the transformer characteristics, such as "160kVA" or "H61_50kVA".

            windings:
                The type of windings such as "Dyn11".

        Returns:
            The constructed transformer characteristics.
        """
        if name == "H61_50kVA":
            return cls(name, windings, 20000, 400, 50 * 1e3, 145, 1.8 / 100, 1350, 4 / 100)
        elif name[-3:] == "kVA":
            try:
                sn = float(name[:-3])
            except ValueError:
                msg = f"The transformer type name does not follow the syntax rule. {name!r} was provided."
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TYPE_NAME_SYNTAX)
            else:
                return cls(name, windings, 20000, 400, sn * 1e3, 460, 2.3 / 100, 2350, 4 / 100)
        else:
            msg = f"The transformer type name does not follow the syntax rule. {name!r} was provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TYPE_NAME_SYNTAX)

    @classmethod
    def from_dict(cls, characteristics: dict[str, Any]):
        """Dict constructor

        Args:
            characteristics:
                A dictionary of the characteristics of the transformer
        """
        type_name = characteristics["name"]
        sn = characteristics["sn"]
        uhv = characteristics["uhv"]  # Phase-to-phase nominal voltages of the high voltages side (V)
        ulv = characteristics["ulv"]  # Phase-to-phase nominal voltages of the low voltages side (V)
        i0 = characteristics["i0"]
        p0 = characteristics["p0"]  # Losses during off-load test (W)
        psc = characteristics["psc"]  # Losses during short circuit test (W)
        vsc = characteristics["vsc"]
        windings = characteristics["type"]  # Winding of the transformer

        return cls(type_name=type_name, windings=windings, uhv=uhv, ulv=ulv, sn=sn, p0=p0, i0=i0, psc=psc, vsc=vsc)

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.type_name,
            "sn": self.sn,
            "uhv": self.uhv,
            "ulv": self.ulv,
            "i0": self.i0,
            "p0": self.p0,
            "psc": self.psc,
            "vsc": self.vsc,
            "type": self.windings,
        }

    def to_zyk(self) -> tuple[complex, complex, float, float]:
        """Compute the transformer characteristics z2, ym, k and orientation mandatory for some models

        Returns:
            * z2 The series impedance of the transformer (Ohms).
            * ym The magnetizing admittance of the transformer (Siemens).
            * k The transformation ratio
            * orientation 1 for direct winding, -1 for reverse winding
        """
        # Extract the windings of the primary and the secondary of the transformer
        winding1, winding2, phase_displacement = TransformerType.extract_windings(self.windings)

        # Off-load test
        # Iron losses resistance (Ohm)
        r_iron = self.uhv**2 / self.p0
        # Magnetizing inductance (Henry) * omega (rad/s)
        if self.i0 * self.sn > self.p0:
            lm_omega = self.uhv**2 / (np.sqrt((self.i0 * self.sn) ** 2 - self.p0**2))
            ym = 1 / r_iron + 1 / (1j * lm_omega)
        else:
            ym = 1 / r_iron

        # Short circuit test
        r2 = self.psc * (self.ulv / self.sn) ** 2
        l2_omega = np.sqrt((self.vsc * self.ulv**2 / self.sn) ** 2 - r2**2)
        z2 = r2 + 1j * l2_omega

        # Change the voltages if the reference voltages is phase to neutral
        uhv = self.uhv
        ulv = self.ulv
        if winding1[0] in ("y", "Y"):
            uhv /= np.sqrt(3.0)
        if winding2[0] in ("y", "Y"):
            ulv /= np.sqrt(3.0)
        if winding1[0] in ("z", "Z"):
            uhv /= 3.0
        if winding2[0] in ("z", "Z"):
            ulv /= 3.0

        if phase_displacement in (5, 6):
            # Reverse winding
            return z2, ym, ulv / uhv, -1.0
        else:
            # Normal winding
            assert phase_displacement in (0, 11)
            return z2, ym, ulv / uhv, 1.0



import logging
from abc import ABC
from collections.abc import Sequence
from typing import Any, Optional

import numpy as np
import shapely.wkt
from pint import Quantity
from shapely.geometry import Point, shape

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models.core import Element, Ground
from roseau.load_flow.utils.json_mixin import JsonMixin
from roseau.load_flow.utils.units import ureg

logger = logging.getLogger(__name__)


class AbstractBus(Element, JsonMixin, ABC):
    """This is an abstract class for all different types of buses."""

    _voltage_source_class: Optional[type["VoltageSource"]] = None
    _bus_class: Optional[type["Bus"]] = None

    def __init__(
        self,
        id: Any,
        n: int,
        potentials: Optional[Sequence[complex]] = None,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """Abstract bus constructor.

        Args:
            id:
                The identifier of the bus.

            n:
                Number of ports ie number of phases.

            potentials:
                List of initial potentials of each phase.

            geometry:
                The geometry of the bus.
        """
        super().__init__(**kwargs)
        self.id = id
        self.n = n

        if potentials is None:
            potentials = np.zeros(n, dtype=complex)
            self.initialized = False
        else:
            if len(potentials) != n:
                msg = f"Incorrect number of potentials: {len(potentials)} instead of {n}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_POTENTIALS_SIZE)
            if isinstance(potentials, Quantity):
                potentials = potentials.m_as("V")
            self.initialized = True
        self.initial_potentials = np.asarray(potentials)
        self.geometry = geometry
        self._potentials = None

    def __str__(self) -> str:
        return f"id={self.id} - n={self.n}"

    @property
    @ureg.wraps("V", None, strict=False)
    def potentials(self) -> np.ndarray:
        """Return the potentials of the element.

        Returns:
            An array of the potentials.
        """
        return self._potentials

    @potentials.setter
    def potentials(self, value: np.ndarray):
        self._potentials = value

    @property
    @ureg.wraps("V", None, strict=False)
    def voltages(self) -> np.ndarray:
        """Return the voltages of the element, as [Van, Vbn, Vcn] for a wye bus, or [Vab, Vbc, Vca] for a delta bus.

        Returns:
            An array of the voltages.
        """
        if self.n == 3:
            return np.asarray(
                self._potentials[1] - self._potentials[0],  # ab
                self._potentials[2] - self._potentials[1],  # bc
                self._potentials[0] - self._potentials[2],  # ca
            )
        else:
            return self._potentials[: self.n - 1] - self._potentials[self.n - 1]  # an, bn, cn

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data, ground):
        if "geometry" not in data:
            geometry = None
        elif isinstance(data["geometry"], str):
            geometry = shapely.wkt.loads(data["geometry"])
        else:
            geometry = shape(data["geometry"])

        potentials = data["potentials"] if "potentials" in data else None
        if data["type"] == "slack":
            v = data["voltages"]
            voltages = [v["va"][0] + 1j * v["va"][1], v["vb"][0] + 1j * v["vb"][1], v["vc"][0] + 1j * v["vc"][1]]
            return cls._voltage_source_class(
                id=data["id"], n=4, ground=ground, source_voltages=voltages, potentials=potentials, geometry=geometry
            )
        else:
            if data["type"] not in ["bus", "bus_neutral"]:
                msg = f"Bad bus type for bus {data['id']}: {data['type']}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_BUS_TYPE)
            if "neutral" in data["type"]:
                bus = cls._bus_class(id=data["id"], n=4, potentials=potentials, geometry=geometry)
            else:
                bus = cls._bus_class(id=data["id"], n=3, potentials=potentials, geometry=geometry)
            return bus


class VoltageSource(AbstractBus):
    """A VoltageSource bus.

    The equations are the following:

    .. math::
        \\left(V_{\\mathrm{a}}-V_{\\mathrm{n}}\\right) &= U_{\\mathrm{a}} \\\\
        \\left(V_{\\mathrm{b}}-V_{\\mathrm{n}}\\right) &= U_{\\mathrm{b}} \\\\
        \\left(V_{\\mathrm{c}}-V_{\\mathrm{n}}\\right) &= U_{\\mathrm{c}}
    """

    def __init__(
        self,
        id: Any,
        n: int,
        ground: Optional[Ground],
        source_voltages: Sequence[complex],
        potentials: Optional[Sequence[complex]] = None,
        geometry: Optional[Point] = None,
        **kwargs,
    ):
        """VoltageSource constructor.

        Args:
            id:
                The identifier of the bus.

            n:
                Number of ports.

            ground:
                The ground to connect the neutral to.

            source_voltages:
                List of the voltages of the source (V).

            potentials:
                List of initial potentials of each phase (V).

            geometry:
                The geometry of the bus.
        """
        super().__init__(
            id=id, n=n, ground=ground, voltages=source_voltages, potentials=potentials, geometry=geometry, **kwargs
        )
        if len(source_voltages) != n - 1:
            msg = f"Incorrect number of voltages: {len(source_voltages)} instead of {n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_VOLTAGES_SIZE)

        if isinstance(source_voltages, Quantity):
            source_voltages = source_voltages.m_as("V")

        if ground is not None:
            ground.connected_elements.append(self)
            self.connected_elements.append(ground)

        self.source_voltages = source_voltages

    @ureg.wraps(None, (None, "V"), strict=False)
    def update_source_voltages(self, source_voltages: Sequence[complex]) -> None:
        """Change the voltages of the source

        Args:
            source_voltages:
                The new voltages.
        """
        if len(source_voltages) != self.n - 1:
            msg = f"Incorrect number of voltages: {len(source_voltages)} instead of {self.n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_VOLTAGES_SIZE)

        self.source_voltages = source_voltages

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        va = self.source_voltages[0]
        vb = self.source_voltages[1]
        vc = self.source_voltages[2]
        res = {
            "id": self.id,
            "type": "slack",
            "loads": [],
            "voltages": {
                "va": [va.real, va.imag],
                "vb": [vb.real, vb.imag],
                "vc": [vc.real, vc.imag],
            },
        }
        if self.geometry is not None:
            res["geometry"] = self.geometry.__geo_interface__
        return res


class Bus(AbstractBus):
    """This class is used for a general purpose bus."""

    def __init__(
        self,
        id: Any,
        n: int,
        potentials: Optional[Sequence[complex]] = None,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """Bus constructor.

        Args:
            id:
                The identifier of the bus.

            n:
                Number of ports.

            potentials:
                List of initial potentials of each phase (V)

            geometry:
                The geometry of the bus.
        """
        super().__init__(id=id, n=n, potentials=potentials, geometry=geometry, **kwargs)

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        bus_type = "bus" if self.n == 3 else "bus_neutral"
        res = {"id": self.id, "type": bus_type, "loads": []}
        if self.geometry is not None:
            res["geometry"] = self.geometry.__geo_interface__
        return res


AbstractBus._voltage_source_class = VoltageSource
AbstractBus._bus_class = Bus



import logging
from abc import ABC
from typing import Any, Optional, TYPE_CHECKING

import numpy as np
import shapely.wkt
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.utils import BranchType
from roseau.load_flow.utils.json_mixin import JsonMixin
from roseau.load_flow.utils.units import ureg

if TYPE_CHECKING:
    from roseau.load_flow.models.buses import AbstractBus
    from roseau.load_flow.models.lines import AbstractLine, Switch
    from roseau.load_flow.models.transformers import AbstractTransformer

logger = logging.getLogger(__name__)


class Element(ABC):
    """An abstract class to describe an element of an Electrical network"""

    def __init__(self, **kwargs):
        self.connected_elements: list[Element] = []

    def disconnect(self):
        """Remove all the connections with the other elements."""
        for element in self.connected_elements:
            element.connected_elements[:] = [e for e in element.connected_elements if e != self]


class PotentialRef(Element):
    """This elements defines the potential reference"""

    def __init__(self, element: Element, **kwargs):
        """Potential reference element constructor, this element will set the origin of the potentials as
        Va + Vb + Vc = 0 for delta elements or Vn = 0 for the others.

        Args:
            element:
                The element to connect to.
        """
        super().__init__(**kwargs)
        self.connected_elements = [element]
        element.connected_elements.append(self)

    @property
    @ureg.wraps("V", None, strict=False)
    def current(self) -> complex:
        """Compute the sum of the currents of the connection associated to the potential reference. This sum should be
        equal to 0 after the load flow.

        Returns:
            The sum of the current of the connection.
        """
        raise NotImplementedError


class Ground(Element):
    """This element defines the ground."""

    def __init__(self, **kwargs):
        """Ground constructor."""
        super().__init__(**kwargs)

    def connect(self, bus: "AbstractBus"):
        """Connect the ground to the bus neutral.

        Args:
            bus:
                The bus to connect to.
        """
        if self not in bus.connected_elements:
            self.connected_elements.append(bus)
            bus.connected_elements.append(self)


class AbstractBranch(Element, JsonMixin):
    """This is an abstract class for all the branches (lines, switches and transformers) of the network."""

    branch_type: BranchType = NotImplemented

    @classmethod
    def _line_class(cls) -> type["AbstractLine"]:
        from roseau.load_flow.models.lines.lines import AbstractLine

        return AbstractLine

    @classmethod
    def _transformer_class(cls) -> type["AbstractTransformer"]:
        from roseau.load_flow.models.transformers.transformers import AbstractTransformer

        return AbstractTransformer

    @classmethod
    def _switch_class(cls) -> type["Switch"]:
        from roseau.load_flow.models.lines.lines import Switch

        return Switch

    def __init__(
        self,
        id: Any,
        n1: int,
        n2: int,
        bus1: "AbstractBus",
        bus2: "AbstractBus",
        geometry: Optional[BaseGeometry] = None,
        **kwargs,
    ) -> None:
        """AbstractBranch constructor.

        Args:
            id:
                The identifier of the branch.

            n1:
                Number of ports in the first extremity of the branch.

            n2:
                Number of ports in the second extremity of the branch.

            bus1:
                Bus to connect to the first extremity of the branch.

            bus2:
                Bus to connect to the second extremity of the branch.

            geometry:
                The geometry of the branch.
        """
        super().__init__(**kwargs)
        self.id = id
        self.n1 = n1
        self.n2 = n2
        self.connected_elements = [bus1, bus2]
        bus1.connected_elements.append(self)
        bus2.connected_elements.append(self)
        self.geometry = geometry
        self._currents = None

    def __str__(self) -> str:
        return f"id={self.id} - n1={self.n1} - n2={self.n2}"

    @property
    @ureg.wraps(("A", "A"), None, strict=False)
    def currents(self) -> tuple[np.ndarray, np.ndarray]:
        """Current accessor

        Returns:
            The complex currents of each phase.
        """
        return self._currents

    @currents.setter
    def currents(self, value: np.ndarray):
        self._currents = value

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, branch, bus1, bus2, ground, line_types, transformer_types, *args):

        if "geometry" not in branch:
            geometry = None
        elif isinstance(branch["geometry"], str):
            geometry = shapely.wkt.loads(branch["geometry"])
        else:
            geometry = shape(branch["geometry"])

        if branch["type"] == "line":
            return cls._line_class().from_dict(
                id=branch["id"],
                bus1=bus1,
                bus2=bus2,
                length=branch["length"],
                line_types=line_types,
                type_name=branch["type_name"],
                ground=ground,
                geometry=geometry,
            )
        elif branch["type"] == "transformer":
            return cls._transformer_class().from_dict(
                id=branch["id"],
                bus1=bus1,
                bus2=bus2,
                type_name=branch["type_name"],
                transformer_types=transformer_types,
                tap=branch["tap"],
                geometry=geometry,
            )
        elif branch["type"] == "switch":
            return cls._switch_class()(id=branch["id"], n=bus1.n, bus1=bus1, bus2=bus2, geometry=geometry)
        else:
            msg = f"Unknown branch type for branch {branch['id']}: {branch['type']}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_BRANCH_TYPE)

    def to_dict(self) -> dict[str, Any]:
        res = {
            "id": self.id,
            "bus1": self.connected_elements[0].id,
            "bus2": self.connected_elements[1].id,
        }
        if self.geometry is not None:
            res["geometry"] = self.geometry.__geo_interface__
        return res



import importlib.metadata

from roseau.load_flow.__about__ import (
    __author__,
    __copyright__,
    __credits__,
    __email__,
    __license__,
    __maintainer__,
    __status__,
    __url__,
)
from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models import (
    AbstractBranch,
    AbstractBus,
    AbstractLine,
    AbstractLoad,
    AbstractTransformer,
    AdmittanceLoad,
    Bus,
    Control,
    DeltaAdmittanceLoad,
    DeltaDeltaTransformer,
    DeltaImpedanceLoad,
    DeltaPowerLoad,
    DeltaWyeTransformer,
    DeltaZigzagTransformer,
    Element,
    FlexibleLoad,
    FlexibleParameter,
    Ground,
    ImpedanceLoad,
    LineCharacteristics,
    PotentialRef,
    PowerLoad,
    Projection,
    ShuntLine,
    SimplifiedLine,
    Switch,
    TransformerCharacteristics,
    VoltageSource,
    WyeDeltaTransformer,
    WyeWyeTransformer,
    WyeZigzagTransformer,
)
from roseau.load_flow.network import ElectricalNetwork

__version__ = importlib.metadata.version("roseau_load_flow")

__all__ = [
    "__author__",
    "__copyright__",
    "__credits__",
    "__email__",
    "__license__",
    "__maintainer__",
    "__status__",
    "__url__",
    "__version__",
    # Electrical Network
    "ElectricalNetwork",
    # Buses
    "AbstractBus",
    "Bus",
    "VoltageSource",
    # Core
    "Element",
    "Ground",
    "PotentialRef",
    "AbstractBranch",
    # Lines
    "Switch",
    "AbstractLine",
    "ShuntLine",
    "SimplifiedLine",
    "LineCharacteristics",
    # Loads
    "AbstractLoad",
    "ImpedanceLoad",
    "PowerLoad",
    "AdmittanceLoad",
    "DeltaPowerLoad",
    "DeltaAdmittanceLoad",
    "DeltaImpedanceLoad",
    "FlexibleLoad",
    "FlexibleParameter",
    "Control",
    "Projection",
    # Transformers
    "AbstractTransformer",
    "WyeWyeTransformer",
    "DeltaWyeTransformer",
    "DeltaDeltaTransformer",
    "WyeDeltaTransformer",
    "WyeZigzagTransformer",
    "DeltaZigzagTransformer",
    "TransformerCharacteristics",
    # Exceptions
    "RoseauLoadFlowException",
    "RoseauLoadFlowExceptionCode",
]



from roseau.load_flow.models.buses import AbstractBus, Bus, VoltageSource
from roseau.load_flow.models.core import AbstractBranch, Element, Ground, PotentialRef
from roseau.load_flow.models.lines import AbstractLine, LineCharacteristics, ShuntLine, SimplifiedLine, Switch
from roseau.load_flow.models.loads import (
    AbstractLoad,
    AdmittanceLoad,
    Control,
    DeltaAdmittanceLoad,
    DeltaImpedanceLoad,
    DeltaPowerLoad,
    FlexibleLoad,
    FlexibleParameter,
    ImpedanceLoad,
    PowerLoad,
    Projection,
)
from roseau.load_flow.models.transformers import (
    AbstractTransformer,
    DeltaDeltaTransformer,
    DeltaWyeTransformer,
    DeltaZigzagTransformer,
    TransformerCharacteristics,
    WyeDeltaTransformer,
    WyeWyeTransformer,
    WyeZigzagTransformer,
)

__all__ = [
    # Core
    "Element",
    "PotentialRef",
    "Ground",
    "AbstractBranch",
    # Buses
    "AbstractBus",
    "Bus",
    "VoltageSource",
    # Lines
    "Switch",
    "AbstractLine",
    "ShuntLine",
    "SimplifiedLine",
    "LineCharacteristics",
    # Loads
    "AbstractLoad",
    "ImpedanceLoad",
    "PowerLoad",
    "AdmittanceLoad",
    "DeltaAdmittanceLoad",
    "DeltaImpedanceLoad",
    "DeltaPowerLoad",
    "FlexibleLoad",
    "FlexibleParameter",
    "Control",
    "Projection",
    # Transformers
    "AbstractTransformer",
    "WyeWyeTransformer",
    "DeltaWyeTransformer",
    "DeltaDeltaTransformer",
    "WyeDeltaTransformer",
    "WyeZigzagTransformer",
    "DeltaZigzagTransformer",
    "TransformerCharacteristics",
]



from abc import ABCMeta, abstractmethod
from typing import Any


class JsonMixin(metaclass=ABCMeta):
    @staticmethod
    @abstractmethod
    def from_dict(*args):
        """Create an element from a dictionary."""
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        """Return the element information as a dictionary format."""
        raise NotImplementedError



import logging
from abc import ABCMeta
from collections.abc import Sequence
from typing import Any, Optional

import numpy as np
from pint import Quantity

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models.buses import AbstractBus
from roseau.load_flow.models.core import Element
from roseau.load_flow.models.loads.flexible_parameters import FlexibleParameter
from roseau.load_flow.utils.json_mixin import JsonMixin
from roseau.load_flow.utils.units import ureg

logger = logging.getLogger(__name__)


class AbstractLoad(Element, JsonMixin, metaclass=ABCMeta):
    """An abstract class to depict a load."""

    _power_load_class: Optional[type["PowerLoad"]] = None
    _delta_power_load_class: Optional[type["DeltaPowerLoad"]] = None
    _impedance_load_class: Optional[type["ImpedanceLoad"]] = None
    _delta_impedance_load_class: Optional[type["DeltaImpedanceLoad"]] = None
    _admittance_load_class: Optional[type["AdmittanceLoad"]] = None
    _delta_admittance_load_class: Optional[type["DeltaAdmittanceLoad"]] = None
    _flexible_load_class: Optional[type["FlexibleLoad"]] = None

    def __init__(self, id: Any, n: int, bus: AbstractBus, **kwargs) -> None:
        """Load constructor.

        Args:
            id:
                The id of the load.

            n:
                Number of ports

            bus:
                Bus to be attached to.
        """
        super().__init__(**kwargs)
        self.connected_elements = [bus]
        bus.connected_elements.append(self)

        self.id = id
        self.n = n
        self.bus = bus
        self._currents = None

    def __str__(self) -> str:
        return f"id={self.id} - n={self.n}"

    @property
    @ureg.wraps("A", None, strict=False)
    def currents(self) -> np.ndarray:
        """Get the actual currents of the load.

        Returns:
            An array containing the actual currents of each phase.
        """
        return self._currents

    @currents.setter
    def currents(self, value: np.ndarray):
        self._currents = value

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data, bus):
        if data["function"] == "flexible":
            return cls._flexible_load_class.from_dict(data=data, bus=bus)
        if "ys" in data["function"] or "ds" in data["function"]:
            s = data["powers"]
            powers = [s["sa"][0] + 1j * s["sa"][1], s["sb"][0] + 1j * s["sb"][1], s["sc"][0] + 1j * s["sc"][1]]
            if "ys" in data["function"]:
                return cls._power_load_class(id=data["id"], n=4, bus=bus, s=powers)
            else:
                return cls._delta_power_load_class(id=data["id"], bus=bus, s=powers)
        elif "yy" in data["function"] or "dy" in data["function"]:
            y = data["admittances"]
            admittances = [y["ya"][0] + 1j * y["ya"][1], y["yb"][0] + 1j * y["yb"][1], y["yc"][0] + 1j * y["yc"][1]]
            if "yy" in data["function"]:
                return cls._admittance_load_class(id=data["id"], n=4, bus=bus, y=admittances)
            else:
                return cls._delta_admittance_load_class(id=data["id"], bus=bus, y=admittances)
        elif "yz" in data["function"] or "dz" in data["function"]:
            z = data["impedances"]
            impedances = [z["za"][0] + 1j * z["za"][1], z["zb"][0] + 1j * z["zb"][1], z["zc"][0] + 1j * z["zc"][1]]
            if "yz" in data["function"]:
                return cls._impedance_load_class(id=data["id"], n=4, bus=bus, z=impedances)
            else:
                return cls._delta_impedance_load_class(id=data["id"], bus=bus, z=impedances)
        else:
            msg = f"Unknown load type for load {data['id']}: {data['function']}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_LOAD_TYPE)


class PowerLoad(AbstractLoad):
    """A constant power load.

    The equations are the following (star loads):

    .. math::
        I_{\\mathrm{abc}} &= \\left(\\frac{S_{\\mathrm{abc}}}{V_{\\mathrm{abc}}-V_{\\mathrm{n}}}\\right)^{\\star} \\\\
        I_{\\mathrm{n}} &= -\\sum_{p\\in\\{\\mathrm{a},\\mathrm{b},\\mathrm{c}\\}}I_{p}

    """

    def __init__(self, id: Any, n: int, bus: AbstractBus, s: Sequence[complex], **kwargs) -> None:
        """PowerLoad constructor.

        Args:
            id:
                The id of the load.

            n:
                Number of ports

            bus:
                Bus to be attached to

            s:
                List of power for each phase (VA).
        """
        super().__init__(id=id, n=n, bus=bus, **kwargs)
        if len(s) != n - 1:
            msg = f"Incorrect number of powers: {len(s)} instead of {n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_SIZE)

        if isinstance(s, Quantity):
            s = s.m_as("VA")
        self.s = s

    @ureg.wraps(None, (None, "VA"), strict=False)
    def update_powers(self, s: Sequence[complex]) -> None:
        """Change the power of the load.

        Args:
            s:
                The new powers to set (VA).
        """
        if len(s) != self.n - 1:
            msg = f"Incorrect number of powers: {len(s)} instead of {self.n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_SIZE)
        self.s = s

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        if self.bus.n == 3:
            load_type = "ys"
        else:
            load_type = "ys_neutral"
        sa, sb, sc = self.s
        return {
            "id": self.id,
            "function": load_type,
            "powers": {
                "sa": [sa.real, sa.imag],
                "sb": [sb.real, sb.imag],
                "sc": [sc.real, sc.imag],
            },
        }


class DeltaPowerLoad(AbstractLoad):
    """A constant power load.

    The equations are the following (delta loads):

    .. math::
        I_{\\mathrm{ab}} &= \\left(\\frac{S_{\\mathrm{ab}}}{V_{\\mathrm{a}}-V_{\\mathrm{b}}}\\right)^{\\star} \\\\
        I_{\\mathrm{bc}} &= \\left(\\frac{S_{\\mathrm{bc}}}{V_{\\mathrm{b}}-V_{\\mathrm{c}}}\\right)^{\\star} \\\\
        I_{\\mathrm{ca}} &= \\left(\\frac{S_{\\mathrm{ca}}}{V_{\\mathrm{c}}-V_{\\mathrm{a}}}\\right)^{\\star}
    """

    def __init__(self, id: Any, bus: AbstractBus, s: Sequence[complex], **kwargs) -> None:
        """PowerLoad constructor.

        Args:
            id:
                The id of the load.

            bus:
                Bus to be attached to

            s:
                List of power for each phase (VA).
        """
        if "n" in kwargs:
            kwargs.pop("n")
        super().__init__(id=id, n=3, bus=bus, **kwargs)
        if len(s) != 3:
            msg = f"Incorrect number of powers: {len(s)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_SIZE)

        if isinstance(s, Quantity):
            s = s.m_as("VA")
        self.s = s

    @ureg.wraps(None, (None, "VA"), strict=False)
    def update_powers(self, s: Sequence[complex]) -> None:
        """Change the power of the load.

        Args:
            s:
                The new powers to set (VA).
        """
        if len(s) != 3:
            msg = f"Incorrect number of powers: {len(s)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_SIZE)
        self.s = s

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        load_type = "ds"
        sa, sb, sc = self.s
        return {
            "id": self.id,
            "function": load_type,
            "powers": {
                "sa": [sa.real, sa.imag],
                "sb": [sb.real, sb.imag],
                "sc": [sc.real, sc.imag],
            },
        }


class AdmittanceLoad(AbstractLoad):
    """A constant admittance load.

    The equations are the following (star loads):

    .. math::
        I_{\\mathrm{abc}} &= Y_{\\mathrm{abc}}\\left(V_{\\mathrm{abc}}-V_{\\mathrm{n}}\\right) \\\\
        I_{\\mathrm{n}} &= -\\sum_{p\\in\\{\\mathrm{a},\\mathrm{b},\\mathrm{c}\\}}I_{p}
    """

    def __init__(self, id: Any, n: int, bus: AbstractBus, y: Sequence[complex], **kwargs) -> None:
        """AdmittanceLoad constructor.

        Args:
            id:
                The id of the load.

            n:
                Number of ports.

            bus:
                Bus to be attached to.

            y:
                List of admittance for each phase (Siemens).
        """
        super().__init__(id=id, n=n, bus=bus, **kwargs)
        if len(y) != n - 1:
            msg = f"Incorrect number of admittance: {len(y)} instead of {n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SIZE)

        if isinstance(y, Quantity):
            y = y.m_as("S")

        self.y = y

    @ureg.wraps(None, (None, "S"), strict=False)
    def update_admittances(self, y: Sequence[complex]) -> None:
        """Change the admittances of the load

        Args:
            y:
                The new admittances to set (Siemens).
        """
        if len(y) != self.n - 1:
            msg = f"Incorrect number of admittances: {len(y)} instead of {self.n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SIZE)
        self.y = y

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        if self.bus.n == 3:
            load_type = "yy"
        else:
            load_type = "yy_neutral"
        ya, yb, yc = self.y
        return {
            "id": self.id,
            "function": load_type,
            "admittances": {
                "ya": [ya.real, ya.imag],
                "yb": [yb.real, yb.imag],
                "yc": [yc.real, yc.imag],
            },
        }


class DeltaAdmittanceLoad(AbstractLoad):
    """A constant admittance load.

    The equations are the following (delta loads):

    .. math::
        I_{\\mathrm{ab}} &= Y_{\\mathrm{ab}}\\left(V_{\\mathrm{a}}-V_{\\mathrm{b}}\\right) \\\\
        I_{\\mathrm{bc}} &= Y_{\\mathrm{bc}}\\left(V_{\\mathrm{b}}-V_{\\mathrm{c}}\\right) \\\\
        I_{\\mathrm{ca}} &= Y_{\\mathrm{ca}}\\left(V_{\\mathrm{c}}-V_{\\mathrm{a}}\\right)
    """

    def __init__(self, id: Any, bus: AbstractBus, y: Sequence[complex], **kwargs) -> None:
        """AdmittanceLoad constructor.

        Args:
            id:
                The id of the load.

            bus:
                Bus to be attached to.

            y:
                List of admittance for each phase (Siemens).
        """
        if "n" in kwargs:
            kwargs.pop("n")
        super().__init__(id=id, n=3, bus=bus, **kwargs)
        if len(y) != 3:
            msg = f"Incorrect number of admittance: {len(y)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SIZE)

        if isinstance(y, Quantity):
            y = y.m_as("S")

        self.y = y

    @ureg.wraps(None, (None, "S"), strict=False)
    def update_admittances(self, y: Sequence[complex]) -> None:
        """Change the admittances of the load

        Args:
            y:
                The new admittances to set (Siemens).
        """
        if len(y) != 3:
            msg = f"Incorrect number of admittances: {len(y)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Y_SIZE)
        self.y = y

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        load_type = "dy"
        ya, yb, yc = self.y
        return {
            "id": self.id,
            "function": load_type,
            "admittances": {
                "ya": [ya.real, ya.imag],
                "yb": [yb.real, yb.imag],
                "yc": [yc.real, yc.imag],
            },
        }


class ImpedanceLoad(AbstractLoad):
    """Constant impedance loads.

    The equations are the same as in the constance admittance load implementation.
    """

    def __init__(self, id: Any, n: int, bus: AbstractBus, z: Sequence[complex], **kwargs) -> None:
        """ImpedanceLoad constructor.

        Args:
            id:
                The id of the load.

            n:
                Number of ports.

            bus:
                Bus to be attached to.

            z:
                List of impedance for each phase (Ohms).
        """
        super().__init__(id=id, n=n, bus=bus, **kwargs)
        if len(z) != n - 1:
            msg = f"Incorrect number of impedance: {len(z)} instead of {n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_SIZE)

        if isinstance(z, Quantity):
            z = z.m_as("ohm")

        for zi in z:
            if np.isclose(zi, 0):
                msg = f"An impedance for load {self.id!r} is null"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_VALUE)

        self.z = z

    @ureg.wraps(None, (None, "ohm"), strict=False)
    def update_impedance(self, z: Sequence[complex]) -> None:
        """Change the admittances of the load

        Args:
            z:
                The new impedance to set (Ohms).
        """
        if len(z) != self.n - 1:
            msg = f"Incorrect number of impedance: {len(z)} instead of {self.n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_SIZE)

        for zi in z:
            if np.isclose(zi, 0):
                msg = f"An impedance for load {self.id!r} is null"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_VALUE)
        self.z = z

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        if self.bus.n == 3:
            load_type = "yz"
        else:
            load_type = "yz_neutral"
        za, zb, zc = self.z
        return {
            "id": self.id,
            "function": load_type,
            "impedances": {
                "za": [za.real, za.imag],
                "zb": [zb.real, zb.imag],
                "zc": [zc.real, zc.imag],
            },
        }


class DeltaImpedanceLoad(AbstractLoad):
    """Constant delta impedance loads.

    The equations are the same as in the constance admittance delta load implementation.
    """

    def __init__(self, id: Any, bus: AbstractBus, z: Sequence[complex], **kwargs) -> None:
        """ImpedanceLoad constructor.

        Args:
            id:
                The id of the load.

            bus:
                Bus to be attached to.

            z:
                List of impedance for each phase (Ohms).
        """
        if "n" in kwargs:
            kwargs.pop("n")
        super().__init__(id=id, n=3, bus=bus, **kwargs)
        if len(z) != 3:
            msg = f"Incorrect number of impedance: {len(z)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_SIZE)

        if isinstance(z, Quantity):
            z = z.m_as("ohm")

        for zi in z:
            if np.isclose(zi, 0):
                msg = f"An impedance for load {self.id!r} is null"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_VALUE)

        self.z = z

    @ureg.wraps(None, (None, "ohm"), strict=False)
    def update_impedance(self, z: Sequence[complex]) -> None:
        """Change the admittances of the load

        Args:
            z:
                The new impedance to set (Ohms).
        """
        if len(z) != 3:
            msg = f"Incorrect number of impedance: {len(z)} instead of 3"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_SIZE)

        for zi in z:
            if np.isclose(zi, 0):
                msg = f"An impedance for load {self.id!r} is null"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_Z_VALUE)
        self.z = z

    #
    # Json Mixin interface
    #
    def to_dict(self) -> dict[str, Any]:
        load_type = "dz"
        za, zb, zc = self.z
        return {
            "id": self.id,
            "function": load_type,
            "impedances": {
                "za": [za.real, za.imag],
                "zb": [zb.real, zb.imag],
                "zc": [zc.real, zc.imag],
            },
        }


class FlexibleLoad(AbstractLoad):
    """A class to depict a flexible load i.e. a load with control."""

    flexible_parameter_class: type[FlexibleParameter] = FlexibleParameter

    def __init__(
        self, id: Any, n: int, bus: AbstractBus, s: Sequence[complex], parameters: list[FlexibleParameter], **kwargs
    ):
        """FlexibleLoad constructor.

        Args:
            id:
                The id of the load.

            n:
                Number of ports.

            bus:
                Bus to be attached to.

            s:
                List of theoretic powers for each phase.

            parameters:
                List of flexible parameters for each phase.
        """
        super().__init__(id=id, n=n, bus=bus, **kwargs)
        if len(s) != n - 1:
            msg = f"Incorrect number of powers: {len(s)} instead of {n - 1}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_SIZE)
        if len(parameters) != n - 1:
            msg = f"Incorrect number of parameters: {len(parameters)} instead of {n}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_PARAMETERS_SIZE)

        if isinstance(s, Quantity):
            s = s.m_as("VA")

        for power, parameter in zip(s, parameters):
            if abs(power) > parameter.s_max and (
                parameter.control_p.type != "constant" or parameter.control_q.type != "constant"
            ):
                msg = f"The power is greater than the parameter s_max for flexible load {self.id!r}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_VALUE)
            if parameter.control_p.type == "p_max_u_production" and power.real > 0:
                msg = f"There is a production control but a positive power for flexible load {self.id!r}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_VALUE)
            if parameter.control_p.type == "p_max_u_consumption" and power.real < 0:
                msg = f"There is a consumption control but a negative power for flexible load {self.id!r}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_VALUE)
            if parameter.control_p.type != "constant" and power.real == 0:
                msg = f"There is a P control but a null active power for flexible load {self.id!r}"
                logger.error(msg)
                raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_S_VALUE)

        self.s = s
        self.parameters = parameters
        self._powers = None

    def update_powers(self, s: Sequence[complex]) -> None:
        """Change the power of the load.

        Args:
            s:
                the new powers to set (Volts).
        """
        self.s = s

    @property
    @ureg.wraps("VA", None, strict=False)
    def powers(self) -> np.ndarray:
        """Compute the actual power consumed by the loads.

        Returns:
            An array containing the actual powers of each phase (VA)
        """
        return self._powers

    @powers.setter
    def powers(self, value: np.ndarray):
        self._powers = value

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data: dict[str, Any], bus: AbstractBus) -> "FlexibleLoad":
        s = data["powers"]
        powers = [s["sa"][0] + 1j * s["sa"][1], s["sb"][0] + 1j * s["sb"][1], s["sc"][0] + 1j * s["sc"][1]]
        parameters = list()
        for parameter in data["parameters"]:
            parameters.append(cls.flexible_parameter_class.from_dict(parameter))
        return cls(id=data["id"], n=4, bus=bus, s=powers, parameters=parameters)

    def to_dict(self) -> dict[str, Any]:
        parameter_res = []
        for parameter in self.parameters:
            parameter_res.append(parameter.to_dict())
        sa, sb, sc = self.s
        return {
            "id": self.id,
            "function": "flexible",
            "powers": {
                "sa": [sa.real, sa.imag],
                "sb": [sb.real, sb.imag],
                "sc": [sc.real, sc.imag],
            },
            "parameters": parameter_res,
        }


AbstractLoad._power_load_class = PowerLoad
AbstractLoad._delta_power_load_class = DeltaPowerLoad
AbstractLoad._impedance_load_class = ImpedanceLoad
AbstractLoad._delta_impedance_load_class = DeltaImpedanceLoad
AbstractLoad._admittance_load_class = AdmittanceLoad
AbstractLoad._delta_admittance_load_class = DeltaAdmittanceLoad
AbstractLoad._flexible_load_class = FlexibleLoad



import logging
from abc import ABC
from typing import Any, Optional

from shapely.geometry import Point

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.models.buses import AbstractBus
from roseau.load_flow.models.core import AbstractBranch
from roseau.load_flow.models.transformers.transformers_characteristics import TransformerCharacteristics
from roseau.load_flow.utils import BranchType, TransformerType

logger = logging.getLogger(__name__)


class AbstractTransformer(AbstractBranch, ABC):
    """An abstract class used for all transformers of this package"""

    branch_type = BranchType.TRANSFORMER
    _dd_class: Optional[type["DeltaDeltaTransformer"]] = None
    _dy_class: Optional[type["DeltaWyeTransformer"]] = None
    _dz_class: Optional[type["DeltaZigzagTransformer"]] = None
    _yd_class: Optional[type["WyeDeltaTransformer"]] = None
    _yy_class: Optional[type["WyeWyeTransformer"]] = None
    _yz_class: Optional[type["WyeZigzagTransformer"]] = None

    def __init__(
        self,
        id: Any,
        n1: int,
        n2: int,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """Transformer constructor.

        Args:
            id:
                The identifier of the transformer.

            n1:
                The number of port of the first extremity.

            n2:
                The number of port of the second extremity.

            bus1:
                Bus to connect the first extremity of the transformer.

            bus2:
                Bus to connect the first extremity of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            transformer_characteristics:
                The characteristics of the transformer.

            geometry:
                The geometry of the transformer.
        """
        if geometry is not None and not isinstance(geometry, Point):
            msg = f"The geometry for a {type(self)} must be a point: {geometry.geom_type} provided."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_GEOMETRY_TYPE)

        if tap > 1.1:
            logger.warning(f"The provided tap {tap:.2f} is higher than 1.1. A good value is between 0.9 and 1.1.")
        if tap < 0.9:
            logger.warning(f"The provided tap {tap:.2f} is lower than 0.9. A good value is between 0.9 and 1.1.")

        super().__init__(n1=n1, n2=n2, bus1=bus1, bus2=bus2, id=id, geometry=geometry, **kwargs)
        self.transformer_characteristics = transformer_characteristics
        self.tap = tap

    @classmethod
    def from_dict(
        cls,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        type_name: str,
        transformer_types: dict[str, TransformerCharacteristics],
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        *args,
    ) -> "AbstractTransformer":
        """Transformer constructor from dict.

        Args:
            transformer_types:
                A dictionary of transformer characteristics by type name.

            type_name:
                The name of the transformer type.

            id:
                The identifier of the transformer.

            bus1:
                Bus to connect to the transformer.

            bus2:
                Bus to connect to the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.

        Returns:
            The constructed transformer.
        """
        transformer_characteristics = transformer_types[type_name]
        winding1, winding2, phase_displacement = TransformerType.extract_windings(transformer_characteristics.windings)
        if "Y" in winding1 and "y" in winding2:
            return cls._yy_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        elif "D" in winding1 and "y" in winding2:
            return cls._dy_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        elif "D" in winding1 and "z" in winding2:
            return cls._dz_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        elif "D" in winding1 and "d" in winding2:
            return cls._dd_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        elif "Y" in winding1 and "d" in winding2:
            return cls._yd_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        elif "Y" in winding1 and "z" in winding2:
            return cls._yz_class(
                id=id,
                bus1=bus1,
                bus2=bus2,
                transformer_characteristics=transformer_characteristics,
                tap=tap,
                geometry=geometry,
            )
        else:
            msg = f"Transformer {transformer_characteristics.windings} is not implemented yet..."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)

    def to_dict(self) -> dict[str, Any]:
        res = super().to_dict()
        res.update(
            {
                "type_name": self.transformer_characteristics.type_name,
                "tap": self.tap,
                "type": "transformer",
            }
        )
        return res

    def update_characteristics(self, transformer_characteristics: TransformerCharacteristics, tap: float = 1.0) -> None:
        """Change the transformer parameters

        Args:
            transformer_characteristics:
                The new transformer characteristics.

            tap:
                The tap of the transformer, for example 1.02.
        """
        windings1 = self.transformer_characteristics.windings
        windings2 = transformer_characteristics.windings
        if windings1 != windings2:
            msg = f"The updated windings changed for transformer {self.id!r}: {windings1} to {windings2}."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)
        if tap > 1.1:
            logger.warning(f"The provided tap {tap:.2f} is higher than 1.1. A good value is between 0.9 and 1.1.")
        if tap < 0.9:
            logger.warning(f"The provided tap {tap:.2f} is lower than 0.9. A good value is between 0.9 and 1.1.")

        self.transformer_characteristics = transformer_characteristics
        self.tap = tap


class WyeWyeTransformer(AbstractTransformer):
    """A class to describe Wye-Wye transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """WyeWyeTransformer constructor

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=4,
            n2=4,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "Y" or transformer_characteristics.winding2[0] != "y":
            msg = f"Bad windings for WyeWyeTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


class DeltaWyeTransformer(AbstractTransformer):
    """A class to describe Delta-Wye transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """DeltaWyeTransformer constructor

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=3,
            n2=4,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "D" or transformer_characteristics.winding2[0] != "y":
            msg = f"Bad windings for DeltaWyeTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


class DeltaDeltaTransformer(AbstractTransformer):
    """A class to describe Delta-Delta transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """DeltaDeltaTransformer constructor

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=3,
            n2=3,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "D" or transformer_characteristics.winding2[0] != "d":
            msg = f"Bad windings for DeltaDeltaTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


class WyeDeltaTransformer(AbstractTransformer):
    """A class to describe Wye-Delta transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """WyeDeltaTransformer

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=4,
            n2=3,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "Y" or transformer_characteristics.winding2[0] != "d":
            msg = f"Bad windings for WyeDeltaTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


class WyeZigzagTransformer(AbstractTransformer):
    """A class to describe Wye-Zigzag transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """WyeZigzagTransformer constructor

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=4,
            n2=4,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "Y" or transformer_characteristics.winding2[0] != "z":
            msg = f"Bad windings for WyeZigzagTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


class DeltaZigzagTransformer(AbstractTransformer):
    """A class to describe Delta-Zigzag transformers."""

    def __init__(
        self,
        id: Any,
        bus1: AbstractBus,
        bus2: AbstractBus,
        transformer_characteristics: TransformerCharacteristics,
        tap: float = 1.0,
        geometry: Optional[Point] = None,
        **kwargs,
    ) -> None:
        """DeltaZigzagTransformer constructor

        Args:
            id:
                The identifier of the transformer.

            bus1:
                bus to connect to the transformer

            bus2:
                bus to connect to the transformer

            transformer_characteristics:
                The characteristics of the transformer.

            tap:
                The tap of the transformer, for example 1.02.

            geometry:
                The geometry of the transformer.
        """
        super().__init__(
            id=id,
            n1=3,
            n2=4,
            bus1=bus1,
            bus2=bus2,
            transformer_characteristics=transformer_characteristics,
            tap=tap,
            geometry=geometry,
            **kwargs,
        )
        if transformer_characteristics.winding1[0] != "D" or transformer_characteristics.winding2[0] != "z":
            msg = f"Bad windings for DeltaZigzagTransformer {self.id!r}: {transformer_characteristics.windings}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)


AbstractTransformer._dd_class = DeltaDeltaTransformer
AbstractTransformer._dy_class = DeltaWyeTransformer
AbstractTransformer._dz_class = DeltaZigzagTransformer
AbstractTransformer._yd_class = WyeDeltaTransformer
AbstractTransformer._yy_class = WyeWyeTransformer
AbstractTransformer._yz_class = WyeZigzagTransformer



from roseau.load_flow.models.loads.flexible_parameters import Control, FlexibleParameter, Projection
from roseau.load_flow.models.loads.loads import (
    AbstractLoad,
    AdmittanceLoad,
    DeltaAdmittanceLoad,
    DeltaImpedanceLoad,
    DeltaPowerLoad,
    FlexibleLoad,
    ImpedanceLoad,
    PowerLoad,
)

__all__ = [
    "AbstractLoad",
    "ImpedanceLoad",
    "PowerLoad",
    "AdmittanceLoad",
    "FlexibleLoad",
    "FlexibleParameter",
    "Control",
    "Projection",
    "DeltaImpedanceLoad",
    "DeltaPowerLoad",
    "DeltaAdmittanceLoad",
]



import logging
from enum import auto, Enum, unique
from typing import Optional

import regex

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode

# The local logger
logger = logging.getLogger(__name__)


@unique
class LineType(Enum):
    """The type of a line."""

    OVERHEAD = auto()
    """The line is an overhead line."""
    UNDERGROUND = auto()
    """The line is an underground line."""
    TWISTED = auto()
    """The line is a twisted line."""

    def __str__(self) -> str:
        """Print a `LineType`

        Returns:
             A printable string of the line type.
        """
        return super().__str__().split(".", 1)[-1].lower()

    @classmethod
    def from_string(cls, string: str) -> "LineType":
        """Convert a string into a LineType

        Args:
            string:
                The string to convert

        Returns:
            The corresponding LineType.
        """
        string = string.lower()
        if string in ("overhead", "aérien", "aerien", "galerie", "a", "o"):
            return cls.OVERHEAD
        elif string in ("underground", "souterrain", "sous-marin", "s", "u"):
            return cls.UNDERGROUND
        elif string in ("twisted", "torsadé", "torsade", "t"):
            return cls.TWISTED
        else:
            msg = f"The string {string!r} can not be converted into a LineType."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_LINE_TYPE)

    #
    # WordingCodeMixin
    #
    def code(self) -> str:
        """The code method is modified to retrieve a code that can be used in line type names.

        Returns:
            The code of the enumerated value.
        """
        if self == LineType.OVERHEAD:
            return "A"
        elif self == LineType.UNDERGROUND:
            return "S"
        elif self == LineType.TWISTED:
            return "T"
        else:  # pragma: no cover
            msg = f"There is code missing here. I do not know the LineType {self!r}."
            logger.error(msg)
            raise NotImplementedError(msg)


@unique
class ConductorType(Enum):
    """The type of conductor."""

    AL = auto()
    """The conductor is in Aluminium."""
    CU = auto()
    """The conductor is in Copper."""
    AM = auto()
    """The conductor is in Almélec."""
    AA = auto()
    """The conductor is in Alu-Acier."""
    LA = auto()
    """The conductor is in Almélec-Acier."""

    def __str__(self) -> str:
        """Print a `ConductorType`

        Returns:
            A printable string of the conductor type.
        """
        if self == ConductorType.AL:
            return "Al"
        elif self == ConductorType.CU:
            return "Cu"
        elif self == ConductorType.AM:
            return "AM"
        elif self == ConductorType.AA:
            return "AA"
        elif self == ConductorType.LA:
            return "LA"
        else:
            s = super().__str__()
            msg = f"The ConductorType {s} is not known..."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_CONDUCTOR_TYPE)

    @classmethod
    def from_string(cls, string: str) -> "ConductorType":
        """Convert a string into a ConductorType

        Args:
            string:
                The string to convert

        Returns:
            The corresponding ConductorType.
        """
        string = string.lower()
        if string == "al":
            return cls.AL
        elif string == "cu":
            return cls.CU
        elif string == "am":
            return cls.AM
        elif string == "aa":
            return cls.AA
        elif string == "la":
            return cls.LA
        else:
            msg = f"The string {string!r} can not be converted into a ConductorType."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_CONDUCTOR_TYPE)

    #
    # WordingCodeMixin
    #
    def code(self) -> str:
        """The code method is modified to retrieve a code that can be used in line type names.

        Returns:
            The code of the enumerated value.
        """
        return self.name.upper()


@unique
class IsolationType(Enum):
    """The type of the isolation for a wire."""

    UNKNOWN = auto()
    """The isolation of the conductor is made with unknown material."""
    HDPE = auto()
    """The isolation of the conductor is made with High-Density PolyEthylene."""
    LDPE = auto()
    """The isolation of the conductor is made with Low-Density PolyEthylene."""
    PEX = auto()
    """The isolation of the conductor is made with Cross-linked polyethylene."""
    EPR = auto()
    """The isolation of the conductor is made with Ethylene-Propylene Rubber."""
    PVC = auto()
    """The isolation of the conductor is made with PolyVinyl Chloride."""

    def __str__(self) -> str:
        """Print a `IsolationType`

        Returns:
            A printable string of the isolation type.
        """
        return super().__str__().split(".", 1)[-1].upper()

    @classmethod
    def from_string(cls, string: str) -> "IsolationType":
        """Convert a string into a IsolationType

        Args:
            string:
                The string to convert

        Returns:
            The corresponding IsolationType.
        """
        if string.lower() in ("", "unknown", "nan"):
            return cls.UNKNOWN
        elif string == "HDPE":
            return cls.HDPE
        elif string == "LDPE":
            return cls.LDPE
        elif string == "PEX":
            return cls.PEX
        elif string == "EPR":
            return cls.EPR
        elif string == "PVC":
            return cls.PVC
        else:
            msg = f"The string {string!r} can not be converted into a IsolationType."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_ISOLATION_TYPE)


@unique
class LineModel(Enum):
    """An enumerated class for the different line models."""

    UNKNOWN = auto()
    """The line is modelled through an unknown model."""
    LV_EXACT = auto()
    """The line is modelled through the position of the wire (with neutral). Some hypothesis limit this model to
            low voltages lines."""
    SYM = auto()
    """The line is modelled using a symmetric model (without neutral)."""
    SYM_NEUTRAL = auto()
    """The line is modelled  using a symmetric model (with neutral)."""
    ZY = auto()
    """The line is modelled using two 3x3 matrices (shunt admittance and line impedance, without neutral)."""
    ZY_NEUTRAL = auto()
    """The line is modelled using two 4x4 matrices (shunt admittance and line impedance, with neutral)."""
    Z = auto()
    """The line is modelled using a single 3x3 matrices (line impedance, without neutral)."""
    Z_NEUTRAL = auto()
    """The line is modelled using a single 4x4 matrices (line impedance, with neutral)."""

    def __str__(self) -> str:
        """Print a `LineModel`

        Returns:
            A printable string of the names of line model.
        """
        return super().__str__().split(".", 1)[-1].lower()

    @classmethod
    def from_string(cls, string: str) -> "LineModel":
        """Convert a string into a LineModel

        Args:
            string:
                The string to convert

        Returns:
            The corresponding LineModel.
        """
        string = string.lower()
        if string == "lv_exact":
            return cls.LV_EXACT
        elif string == "sym":
            return cls.SYM
        elif string == "sym_neutral":
            return cls.SYM_NEUTRAL
        elif string == "zy":
            return cls.ZY
        elif string == "zy_neutral":
            return cls.ZY_NEUTRAL
        elif string == "z":
            return cls.Z
        elif string == "z_neutral":
            return cls.Z_NEUTRAL
        elif string in ("", "nan", "unknown"):
            return cls.UNKNOWN
        else:
            msg = f"The string {string!r} can not be converted into a LineModel."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_LINE_MODEL)

    @classmethod
    def with_neutral(cls) -> tuple["LineModel", ...]:
        """Give the tuple of the line models which support the neutral.

        Returns:
            The tuple of line models which support the neutral.
        """
        return cls.SYM_NEUTRAL, cls.Z_NEUTRAL, cls.ZY_NEUTRAL, cls.LV_EXACT

    @classmethod
    def without_neutral(cls) -> tuple["LineModel", ...]:
        """Give the tuple of the line models which do not support the neutral.

        Returns:
            The tuple of line models which do not support the neutral.
        """
        return cls.SYM, cls.Z, cls.ZY

    @classmethod
    def with_shunt(cls) -> tuple["LineModel", ...]:
        """Give the tuple of the line models which support the shunt.

        Returns:
            The tuple of line models which supports the shunt.
        """
        return cls.LV_EXACT, cls.SYM, cls.SYM_NEUTRAL, cls.ZY, cls.ZY_NEUTRAL

    @classmethod
    def without_shunt(cls) -> tuple["LineModel", ...]:
        """Give the tuple of the line models which do not support the shunt.

        Returns:
            The tuple of line models which do not support the shunt.
        """
        return cls.Z_NEUTRAL, cls.Z


@unique
class BranchType(Enum):
    """The type of 'line' in a network."""

    LINE = auto()
    """The branch is a regular line."""
    TRANSFORMER = auto()
    """The branch is a regular transformer."""
    SWITCH = auto()
    """The branch is a regular switch."""

    def __str__(self) -> str:
        """Print a `BranchType`.

        Returns:
            A printable string of the connection type.
        """
        return super().__str__().split(".", 1)[-1].lower()

    @classmethod
    def from_string(cls, string: str) -> "BranchType":
        """Convert a string into a ConnectionType

        Args:
            string:
                The string to convert

        Returns:
            The corresponding ConnectionType.
        """
        if string == "line":
            return cls.LINE
        elif string == "transformer":
            return cls.TRANSFORMER
        elif string == "switch":
            return cls.SWITCH
        else:
            msg = f"The string {string!r} can not be converted into a BranchType."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_BRANCH_TYPE)


EXTRACT_WINDINGS_RE: regex.Regex = regex.compile(
    "(?(DEFINE)(?P<y_winding>yn?)(?P<d_winding>d)(?P<z_winding>zn?)(?P<p_set_1>[06])"
    "(?P<p_set_2>5|11))"
    ""
    "(?|(?P<w1>(?&y_winding))(?P<w2>(?&y_winding))(?P<p>(?&p_set_1)?)"  # yy
    "|(?P<w1>(?&y_winding))(?P<w2>(?&d_winding))(?P<p>(?&p_set_2)?)"  # yd
    "|(?P<w1>(?&y_winding))(?P<w2>(?&z_winding))(?P<p>(?&p_set_2)?)"  # yz
    "|(?P<w1>(?&d_winding))(?P<w2>(?&z_winding))(?P<p>(?&p_set_1)?)"  # dz
    "|(?P<w1>(?&d_winding))(?P<w2>(?&y_winding))(?P<p>(?&p_set_2)?)"  # dy
    "|(?P<w1>(?&d_winding))(?P<w2>(?&d_winding))(?P<p>(?&p_set_1)?))",  # dd
    regex.IGNORECASE,
)
"""The pattern to extract the winding of the primary and of the secondary of the transformer."""


@unique
class TransformerType(Enum):
    """The type of transformer."""

    Yy = auto()
    """A Wye-Wye transformer without neutral connected to the rest of the network."""
    YNy = auto()
    """A Wye-Wye transformer with a neutral connected to the network on the first winding."""
    YNyn = auto()
    """A Wye-Wye transformer with a neutral connected to the network on the two windings."""
    Yyn = auto()
    """A Wye-Wye transformer with a neutral connected to the network on the second winding."""
    Dz = auto()
    """A Delta-Zigzag transformer without neutral connected to the rest of the network."""
    Dzn = auto()
    """A Delta-Zigzag transformer with a neutral connected to the network on the second winding."""
    Dy = auto()
    """A Delta-Wye transformer without neutral connected to the rest of the network."""
    Dyn = auto()
    """A Delta-Wye transformer with a neutral connected to the network on the second winding."""
    Yz = auto()
    """A Wye-Zigzag transformer without neutral connected to the rest of the network."""
    YNz = auto()
    """A Wye-Zigzag transformer with a neutral connected to the network on the first winding."""
    YNzn = auto()
    """A Wye-Zigzag transformer with a neutral connected to the network on the two windings."""
    Yzn = auto()
    """A Wye-Zigzag transformer with a neutral connected to the network on the second winding."""
    Yd = auto()
    """A Wye-Delta transformer without neutral connected to the rest of the network."""
    YNd = auto()
    """A Wye-Delta transformer with a neutral connected to the network on the first winding."""
    Dd = auto()
    """A Delta-Delta transformer without neutral connected to the rest of the network."""

    def __str__(self) -> str:
        """Print a `TransformerType`

        Returns:
            A printable string of the transformer type.
        """
        return super().__str__().split(".", 1)[-1]

    @classmethod
    def from_string(cls, string: str) -> "TransformerType":
        """Convert a string into a TransformerType

        Args:
            string:
                The string to convert

        Returns:
            The corresponding TransformerType
        """
        winding1, winding2, phase_displacement = TransformerType.extract_windings(string=string)
        try:
            return getattr(cls, f"{winding1}{winding2}")
        except AttributeError:
            msg = f"The string {string!r} can not be converted into a TransformerType."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_TYPE)

    @property
    def windings(self) -> tuple[str, str]:
        """Retrieve the windings and the phase displacement of the current

        Returns:
            The high voltages winding and the low voltages winding.
        """
        winding1, winding2, phase_displacement = self.extract_windings(str(self))
        return winding1, winding2

    @classmethod
    def validate_windings(cls, string: str) -> bool:
        """Validate the windings of the high and low voltages sides

        Args:
            string:
                A string depicting a winding

        Returns:
            True if the provided string corresponds to valid transformer windings.
        """
        try:
            match: regex.regex.Match = EXTRACT_WINDINGS_RE.fullmatch(string=string)
            return bool(match) and bool(match.group("p"))
        except RoseauLoadFlowException:
            return False

    @classmethod
    def extract_windings(cls, string: str) -> tuple[str, str, Optional[int]]:
        """Extract the windings of the high and low voltages sides

        Args:
            string:
                A string depicting a winding

        Returns:
            The high voltages winding, the low voltages winding, and the phase displacement.
        """
        match: regex.regex.Match = EXTRACT_WINDINGS_RE.fullmatch(string=string)
        if match:
            groups = match.groupdict()
            winding1, winding2, phase_displacement = groups["w1"], groups["w2"], groups["p"]
            if phase_displacement:
                return winding1.upper(), winding2.lower(), int(phase_displacement)
            else:
                return winding1.upper(), winding2.lower(), None
        else:
            msg = f"Transformer windings can not be extracted from the string {string!r}."
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_TRANSFORMER_WINDINGS)



import logging
from typing import Any

from roseau.load_flow.exceptions import RoseauLoadFlowException, RoseauLoadFlowExceptionCode
from roseau.load_flow.utils.units import ureg

logger = logging.getLogger(__name__)


class Control:
    """A class to store the important values of a control."""

    DEFAULT_ALPHA: float = 200.0

    @ureg.wraps(None, (None, None, "V", "V", "V", "V", None), strict=False)
    def __init__(self, type: str, u_min: float, u_down: float, u_up: float, u_max: float, alpha: float = DEFAULT_ALPHA):
        """Control constructor

        Args:
            type:
                The type of the control:
                  * "constant" if no control;
                  * "p_max_u_production" to control the pmax of the load based on the voltage (production);
                  * "p_max_u_consumption" to control the pmax of the load based on the voltage (consumption);
                  * "q_u" to control the reactive power using the voltage.

            u_min:
                The minimum voltage i.e. the one the control reached the maximum action.

            u_down:
                The voltage which starts to trigger the control (lower value).

            u_up:
                The voltage  which starts to trigger the control (upper value).

            u_max:
                The maximum voltage i.e. the one the control reached its maximum action.

            alpha:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.
        """
        self.type = type
        self.u_min = u_min
        self.u_down = u_down
        self.u_up = u_up
        self.u_max = u_max
        self.alpha = alpha

    @classmethod
    def constant(cls) -> "Control":
        """Retrieve a constant control."""
        return cls(type="constant", u_min=0.0, u_down=0.0, u_up=0.0, u_max=0.0)

    @classmethod
    @ureg.wraps(None, (None, "V", "V", None), strict=False)
    def p_max_u_production(cls, u_up: float, u_max: float, alpha: float = DEFAULT_ALPHA) -> "Control":
        """Retrieve a control of the type "p_max_u_production" using the provided parameters.

        Args:
            u_up:
                The voltage  which starts to trigger the control (upper value).

            u_max:
                The maximum voltage i.e. the one the control reached its maximum action.

            alpha:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

        Returns:
            The control using the provided parameters.

        .. image:: /_static/Control_PU_Prod.png
            :width: 600
            :align: center
        """
        assert u_up < u_max
        return cls(type="p_max_u_production", u_min=0.0, u_down=0.0, u_up=u_up, u_max=u_max, alpha=alpha)

    @classmethod
    @ureg.wraps(None, (None, "V", "V", None), strict=False)
    def p_max_u_consumption(cls, u_min: float, u_down: float, alpha: float = DEFAULT_ALPHA) -> "Control":
        """Retrieve a control of the type "p_max_u_consumption" using the provided parameters.

        Args:
            u_min:
                The minimum voltage i.e. the one the control reached the maximum action.

            u_down:
                The voltage which starts to trigger the control (lower value).

            alpha:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

        Returns:
            The control using the provided parameters.

        .. image:: /_static/Control_PU_Cons.png
            :width: 600
            :align: center
        """
        assert u_min < u_down
        return cls(type="p_max_u_consumption", u_min=u_min, u_down=u_down, u_up=0.0, u_max=0.0, alpha=alpha)

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "V", "V", None), strict=False)
    def q_u(cls, u_min: float, u_down: float, u_up: float, u_max: float, alpha: float = DEFAULT_ALPHA) -> "Control":
        """Retrieve a control of the type "q_u" using the provided parameters.

        Args:
            u_min:
                The minimum voltage i.e. the one the control reached the maximum action.

            u_down:
                The voltage which starts to trigger the control (lower value).

            u_up:
                The voltage  which starts to trigger the control (upper value).

            u_max:
                The maximum voltage i.e. the one the control reached its maximum action.

            alpha:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

        Returns:
            The control using the provided parameters.

        .. image:: /_static/Control_QU.png
            :width: 600
            :align: center
        """
        assert u_min < u_down < u_up < u_max
        return cls(type="q_u", u_min=u_min, u_down=u_down, u_up=u_up, u_max=u_max, alpha=alpha)

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Control":
        alpha = data["alpha"] if "alpha" in data else cls.DEFAULT_ALPHA
        if data["type"] == "constant":
            return cls.constant()
        elif data["type"] == "p_max_u_production":
            return cls.p_max_u_production(u_up=data["u_up"], u_max=data["u_max"], alpha=alpha)
        elif data["type"] == "p_max_u_consumption":
            return cls.p_max_u_consumption(u_min=data["u_min"], u_down=data["u_down"], alpha=alpha)
        elif data["type"] == "q_u":
            return cls.q_u(
                u_min=data["u_min"], u_down=data["u_down"], u_up=data["u_up"], u_max=data["u_max"], alpha=alpha
            )
        else:
            msg = f"Unsupported control type {data['type']}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_CONTROL_TYPE)

    def to_dict(self) -> dict[str, Any]:
        if self.type == "constant":
            return {"type": "constant"}
        elif self.type == "p_max_u_production":
            return {"type": "p_max_u_production", "u_up": self.u_up, "u_max": self.u_max, "alpha": self.alpha}
        elif self.type == "p_max_u_consumption":
            return {"type": "p_max_u_consumption", "u_min": self.u_min, "u_down": self.u_down, "alpha": self.alpha}
        elif self.type == "q_u":
            return {
                "type": "q_u",
                "u_min": self.u_min,
                "u_down": self.u_down,
                "u_up": self.u_up,
                "u_max": self.u_max,
                "alpha": self.alpha,
            }
        else:
            msg = f"Unsupported control type {self.type!r}"
            logger.error(msg)
            raise RoseauLoadFlowException(msg=msg, code=RoseauLoadFlowExceptionCode.BAD_CONTROL_TYPE)


class Projection:
    """This class defines the projection on the feasible circle for a flexible load."""

    DEFAULT_ALPHA: float = 100.0
    DEFAULT_EPSILON: float = 0.01

    def __init__(self, type: str, alpha: float = DEFAULT_ALPHA, epsilon: float = DEFAULT_EPSILON):
        """Projection constructor.

        Args:
            type:
                The type of the projection. It can be:
                  * "euclidean" for an Euclidean projection on the feasible space;
                  * "keep_p" for maintaining a constant P;
                  * "keep_q" for maintaining a constant Q.

            alpha:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

                .. math::
                    \\sqrt{S}=\\sqrt{\varepsilon \times \\exp\\left(\frac{-{|S|}^2}{\varepsilon}\right) + {|S|}^2}
        """
        self.type = type
        self.alpha = alpha
        self.epsilon = epsilon

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "Projection":
        alpha = data["alpha"] if "alpha" in data else cls.DEFAULT_ALPHA
        epsilon = data["epsilon"] if "epsilon" in data else cls.DEFAULT_EPSILON
        return cls(type=data["type"], alpha=alpha, epsilon=epsilon)

    def to_dict(self) -> dict[str, Any]:
        return {
            "type": self.type,
            "alpha": self.alpha,
            "epsilon": self.epsilon,
        }


class FlexibleParameter:
    """This class stores the required data to make a flexible parameter."""

    control_class: type[Control] = Control
    projection_class: type[Projection] = Projection

    @ureg.wraps(None, (None, None, None, None, "VA"), strict=False)
    def __init__(self, control_p: Control, control_q: Control, projection: Projection, s_max: float) -> None:
        """FlexibleParameter constructor.

        Args:
            control_p:
                The control to apply on the active power.

            control_q:
                The control to apply on the reactive power.

            projection:
                The projection to use to have a feasible result.

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.
        """
        self.control_p = control_p
        self.control_q = control_q
        self.projection = projection
        self.s_max = s_max

    @classmethod
    def constant(cls) -> "FlexibleParameter":
        """Retrieve a constant control i.e. no control at all. It is an equivalent of the constant power load."""
        return cls(
            control_p=cls.control_class.constant(),
            control_q=cls.control_class.constant(),
            projection=cls.projection_class(type="euclidean"),
            s_max=1.0,
        )

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "VA", None, None, None), strict=False)
    def p_max_u_production(
        cls,
        u_up: float,
        u_max: float,
        s_max: float,
        alpha_control: float = Control.DEFAULT_ALPHA,
        alpha_proj: float = Projection.DEFAULT_ALPHA,
        epsilon_proj: float = Projection.DEFAULT_EPSILON,
    ) -> "FlexibleParameter":
        """Class method to build a "p_max_u_production" flexible parameter.

        Args:
            u_up:
                The voltage  which starts to trigger the control (upper value).

            u_max:
                The maximum voltage i.e. the one the control reached its maximum action.

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.

            alpha_control:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

            alpha_proj:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon_proj:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

        Returns:
            The flexible parameter which depicts "p_max_p_u_production" control on the production active power with
            a Euclidean projection.
        """
        control_p = cls.control_class.p_max_u_production(u_up=u_up, u_max=u_max, alpha=alpha_control)
        return cls(
            control_p=control_p,
            control_q=cls.control_class.constant(),
            projection=cls.projection_class(type="euclidean", alpha=alpha_proj, epsilon=epsilon_proj),
            s_max=s_max,
        )

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "VA", None, None, None), strict=False)
    def p_max_u_consumption(
        cls,
        u_min: float,
        u_down: float,
        s_max: float,
        alpha_control: float = Control.DEFAULT_ALPHA,
        alpha_proj: float = Projection.DEFAULT_ALPHA,
        epsilon_proj: float = Projection.DEFAULT_EPSILON,
    ) -> "FlexibleParameter":
        """Class method to build a "p_max_u_consumption" flexible parameter.

        Args:
            u_min:
                The minimum voltage i.e. the one the control reached the maximum action.

            u_down:
                The voltage which starts to trigger the control (lower value).

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.

            alpha_control:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

            alpha_proj:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon_proj:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

        Returns:
            The flexible parameter which depicts "p_max_p_u_consumption" control on the consumption active power with
            a Euclidean projection.
        """
        control_p = cls.control_class.p_max_u_consumption(u_min=u_min, u_down=u_down, alpha=alpha_control)
        return cls(
            control_p=control_p,
            control_q=cls.control_class.constant(),
            projection=cls.projection_class("euclidean", alpha=alpha_proj, epsilon=epsilon_proj),
            s_max=s_max,
        )

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "V", "V", "VA", None, None, None), strict=False)
    def q_u(
        cls,
        u_min: float,
        u_down: float,
        u_up: float,
        u_max: float,
        s_max: float,
        alpha_control: float = Control.DEFAULT_ALPHA,
        alpha_proj: float = Projection.DEFAULT_ALPHA,
        epsilon_proj: float = Projection.DEFAULT_EPSILON,
    ) -> "FlexibleParameter":
        """Class method to build a "q_u" flexible parameter.

        Args:
            u_min:
                The minimum voltage i.e. the one the control reached the maximum action.

            u_down:
                The voltage which starts to trigger the control (lower value).

            u_up:
                The voltage  which starts to trigger the control (upper value).

            u_max:
                The maximum voltage i.e. the one the control reached its maximum action.

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.

            alpha_control:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

            alpha_proj:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon_proj:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

        Returns:
            The flexible parameter which depicts "q_u" control on the reactive power with a Euclidean projection.
        """
        control_q = cls.control_class.q_u(u_min=u_min, u_down=u_down, u_up=u_up, u_max=u_max, alpha=alpha_control)
        return cls(
            control_p=cls.control_class.constant(),
            control_q=control_q,
            projection=cls.projection_class(type="euclidean", alpha=alpha_proj, epsilon=epsilon_proj),
            s_max=s_max,
        )

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "V", "V", "V", "V", "VA", None, None, None), strict=False)
    def pq_u_production(
        cls,
        up_up: float,
        up_max: float,
        uq_min: float,
        uq_down: float,
        uq_up: float,
        uq_max: float,
        s_max: float,
        alpha_control=Control.DEFAULT_ALPHA,
        alpha_proj=Projection.DEFAULT_ALPHA,
        epsilon_proj=Projection.DEFAULT_EPSILON,
    ) -> "FlexibleParameter":
        """Class method to build a "pq_u_production" flexible parameter.

        Args:
            up_up:
                The voltage  which starts to trigger the control on the production (upper value).

            up_max:
                The maximum voltage i.e. the one the control (of production) reached its maximum action.

            uq_min:
                The minimum voltage i.e. the one the control reached the maximum action (reactive power control)

            uq_down:
                The voltage which starts to trigger the reactive power control (lower value).

            uq_up:
                The voltage  which starts to trigger the reactive power control (upper value).

            uq_max:
                The maximum voltage i.e. the one the reactive power control reached its maximum action.

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.

            alpha_control:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

            alpha_proj:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon_proj:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

        Returns:
            The flexible parameter which depicts a control on the active power in production and a reactive
            power control with a Euclidean projection.
        """
        control_p = cls.control_class.p_max_u_production(u_up=up_up, u_max=up_max, alpha=alpha_control)
        control_q = cls.control_class.q_u(u_min=uq_min, u_down=uq_down, u_up=uq_up, u_max=uq_max, alpha=alpha_control)
        return cls(
            control_p=control_p,
            control_q=control_q,
            projection=cls.projection_class(type="euclidean", alpha=alpha_proj, epsilon=epsilon_proj),
            s_max=s_max,
        )

    @classmethod
    @ureg.wraps(None, (None, "V", "V", "V", "V", "V", "V", "VA", None, None, None), strict=False)
    def pq_u_consumption(
        cls,
        up_min: float,
        up_down: float,
        uq_min: float,
        uq_down: float,
        uq_up: float,
        uq_max: float,
        s_max: float,
        alpha_control: float = Control.DEFAULT_ALPHA,
        alpha_proj: float = Projection.DEFAULT_ALPHA,
        epsilon_proj: float = Projection.DEFAULT_EPSILON,
    ) -> "FlexibleParameter":
        """Class method to build a "pq_u_consumption" flexible parameter.

        Args:
            up_min:
                The minimum voltage i.e. the one the active power control reached the maximum action.

            up_down:
                The voltage which starts to trigger the active power control (lower value).

            uq_min:
                The minimum voltage i.e. the one the control reached the maximum action (reactive power control)

            uq_down:
                The voltage which starts to trigger the reactive power control (lower value).

            uq_up:
                The voltage  which starts to trigger the reactive power control (upper value).

            uq_max:
                The maximum voltage i.e. the one the reactive power control reached its maximum action.

            s_max:
                The apparent power of the flexible load (VA). It is the radius of the feasible circle.

            alpha_control:
                An approximation factor used by the family function (soft clip). The greater, the closer the function
                are from the non-differentiable function.

            alpha_proj:
                This value is used to make soft sign function and to build a soft projection function.

            epsilon_proj:
                This value is used to make a smooth sqrt function. It is only used in the Euclidean projection.

        Returns:
            The flexible parameter which depicts a control on the active power in consumption and a reactive
            power control with a Euclidean projection.
        """
        control_p = cls.control_class.p_max_u_consumption(u_min=up_min, u_down=up_down, alpha=alpha_control)
        control_q = cls.control_class.q_u(u_min=uq_min, u_down=uq_down, u_up=uq_up, u_max=uq_max, alpha=alpha_control)
        return cls(
            control_p=control_p,
            control_q=control_q,
            projection=cls.projection_class(type="euclidean", alpha=alpha_proj, epsilon=epsilon_proj),
            s_max=s_max,
        )

    #
    # Json Mixin interface
    #
    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> "FlexibleParameter":
        control_p = cls.control_class.from_dict(data["control_p"])
        control_q = cls.control_class.from_dict(data["control_q"])
        projection = cls.projection_class.from_dict(data["projection"])
        return cls(control_p=control_p, control_q=control_q, projection=projection, s_max=data["s_max"])

    def to_dict(self) -> dict[str, Any]:
        return {
            "control_p": self.control_p.to_dict(),
            "control_q": self.control_q.to_dict(),
            "projection": self.projection.to_dict(),
            "s_max": self.s_max,
        }



import unicodedata
from enum import auto, Enum
from typing import Union


class RoseauLoadFlowExceptionCode(Enum):
    """An abstract class that will be used in every Roseau Packages."""

    # Generic
    BAD_GEOMETRY_TYPE = auto()

    # Buses
    DUPLICATE_BUS_ID = auto()
    BAD_BUS_TYPE = auto()
    BAD_POTENTIALS_SIZE = auto()
    BAD_VOLTAGES_SIZE = auto()

    # Branches
    DUPLICATE_BRANCH_ID = auto()
    BAD_BRANCH_TYPE = auto()
    BAD_Z_LINE_SHAPE = auto()
    BAD_Y_SHUNT_SHAPE = auto()
    BAD_LINE_MODEL = auto()
    BAD_LINE_TYPE = auto()
    BAD_CONDUCTOR_TYPE = auto()
    BAD_ISOLATION_TYPE = auto()
    BAD_Z_LINE_VALUE = auto()
    BAD_Y_SHUNT_VALUE = auto()
    BAD_TRANSFORMER_WINDINGS = auto()
    BAD_TRANSFORMER_TYPE = auto()
    BAD_TRANSFORMER_VOLTAGES = auto()
    BAD_TRANSFORMER_PARAMETERS = auto()
    BAD_TYPE_NAME_SYNTAX = auto()

    # Control
    BAD_CONTROL_TYPE = auto()

    # Load
    DUPLICATE_LOAD_ID = auto()
    BAD_LOAD_TYPE = auto()
    BAD_Y_SIZE = auto()
    BAD_Z_SIZE = auto()
    BAD_Z_VALUE = auto()
    BAD_S_SIZE = auto()
    BAD_S_VALUE = auto()
    BAD_PARAMETERS_SIZE = auto()

    # Network
    BAD_VOLTAGES_SOURCES_CONNECTION = auto()
    SWITCHES_LOOP = auto()
    NO_POTENTIAL_REFERENCE = auto()
    SEVERAL_POTENTIAL_REFERENCE = auto()
    UNKNOWN_ELEMENT = auto()
    NO_VOLTAGE_SOURCE = auto()
    BAD_ELEMENT_OBJECT = auto()
    DISCONNECTED_ELEMENT = auto()
    BAD_ELEMENT_ID = auto()
    NO_LOAD_FLOW_CONVERGENCE = auto()
    BAD_REQUEST = auto()

    # DGS export
    DGS_BAD_PHASE_TECHNOLOGY = auto()
    DGS_BAD_PHASE_NUMBER = auto()

    # JSON export
    JSON_LINE_CHARACTERISTICS_DUPLICATES = auto()
    JSON_TRANSFORMER_CHARACTERISTICS_DUPLICATES = auto()

    @classmethod
    def package_name(cls) -> str:
        return "roseau.load_flow"

    def __str__(self) -> str:
        return f"{self.package_name()}.{self.name}".lower()

    def __eq__(self, other) -> bool:
        if isinstance(other, str):
            return other.lower() == str(self).lower()
        return super().__eq__(other)

    @classmethod
    def from_string(cls, string: Union[str, "RoseauLoadFlowExceptionCode"]) -> "RoseauLoadFlowExceptionCode":
        """A method to convert a string into an error code enumerated type.

        Args:
            string:
                The string depicted the error code. If a good element is given

        Returns:
            The enumerated type value corresponding with `string`.
        """
        if isinstance(string, cls):
            return string
        elif isinstance(string, str):
            pass
        else:
            string = str(string)

        # Withdraw accents and make lowercase
        string = unicodedata.normalize("NFKD", string.lower()).encode("ASCII", "ignore").decode()

        # Withdraw the package prefix (e.g. roseau.core)
        error_str = string.removeprefix(f"{cls.package_name()}.")

        # Get the value of this string
        return cls[error_str.upper()]


class RoseauLoadFlowException(Exception):
    """A base exception for this repository"""

    def __init__(self, msg: str, code: RoseauLoadFlowExceptionCode, *args) -> None:
        """Constructor for RoseauCoreException

        Args:
            msg:
                A message in English.

            code:
                The code related to this exception.
        """
        super().__init__(msg, code, *args)
        self.msg = msg
        self.code = code

    def __str__(self) -> str:
        return str(self.code)



from roseau.load_flow.models.lines.lines import AbstractLine, LineCharacteristics, ShuntLine, SimplifiedLine, Switch

__all__ = ["Switch", "AbstractLine", "ShuntLine", "SimplifiedLine", "LineCharacteristics"]
